package org.ntnu.pythia.engine;

import java.awt.Rectangle;
import java.awt.geom.Point2D;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;

import javax.xml.bind.JAXBException;

import org.lwjgl.input.Keyboard;
import org.ntnu.pythia.engine.components.CodeEditorDialog;
import org.ntnu.pythia.engine.components.GameMenu;
import org.ntnu.pythia.engine.components.InputEventListener;
import org.ntnu.pythia.engine.components.MainMenu;
import org.ntnu.pythia.engine.components.MessageBox;
import org.ntnu.pythia.engine.components.RootPane;
import org.ntnu.pythia.game.CollisionWorld;
import org.ntnu.pythia.game.GameObject;
import org.ntnu.pythia.game.LevelObject;
import org.ntnu.pythia.game.MusicObject;
import org.ntnu.pythia.game.PlayerObject;
import org.ntnu.pythia.game.behaviour.CompositeObject;
import org.ntnu.pythia.game.level.Level;
import org.ntnu.pythia.game.level.LevelFactory;
import org.ntnu.pythia.game.level.LevelLayout;
import org.ntnu.pythia.game.level.SimpleObject;
import org.ntnu.pythia.python.unit.Tester;

import de.matthiasmann.twl.Event;
import de.matthiasmann.twl.GUI;
import de.matthiasmann.twl.renderer.lwjgl.LWJGLRenderer;
import de.matthiasmann.twl.theme.ThemeManager;

/**
 * Main game logic
 * 
 * @author egr
 * 
 */
public class Game implements InputEventListener {
	
	public static final int GAME_STATE_INTRO = 0;
	public static final int GAME_STATE_MAIN_MENU = 1;
	public static final int GAME_STATE_IN_GAME = 2;
	public static final int GAME_STATE_GAME_MENU = 3;
	public static final int GAME_STATE_TRANSITION = 4;
	public static final int GAME_STATE_GAME_OVER = 5;
	public static final int GAME_STATE_GAME_END = 6;
	
	private int state = GAME_STATE_INTRO;
	private int next_state;
	
	public float transition_progress = 0f;
	
	

	private Camera mCamera;
	private PlayerObject mPlayer;
	private LevelObject mLevelObject;
	private final GameWorld gameWorld = new GameWorld(this);
	Level mLevel = null;

	Random generator = new Random();

	private GUI gui;
	private RootPane root;
	private ThemeManager theme;
	private Rectangle view = new Rectangle();
	private MainMenu mainMenu;
	private GameMenu gameMenu;

	private CollisionWorld cWorld = new CollisionWorld();

	private List<GameObject> mGameObjects = new ArrayList<GameObject>();
	List<GameObject> renderObjects = new ArrayList<GameObject>(50);
	private Map<String, GameObject> mGameObjectLookUpMap = Collections
			.synchronizedMap(new HashMap<String, GameObject>());

	private boolean loadNextLevel = false;
	private String nextLevel;

	private final Comparator<GameObject> mGameObjectComparator = new GameObjectComparator();

	/**
	 * Nothing to see here, move along
	 */
	public Game() {
		root = new RootPane();
		root.setTheme("");
		root.setEventListener(this);
	}

	public void loadLevel(String name) {
		loadNextLevel = true;
		nextLevel = name;
	}

	/**
	 * Internal level loading, avoid check if the game is in progress
	 * 
	 * @param name
	 */
	private void loadNextLevel(String name) {
		// Construct a level
		mGameObjects.clear();
		App.globals.getParser().clear();
		mGameObjectLookUpMap.clear();
		LevelLayout mLevelLayout = null;
		try {
			if(App.options.isExternal())
				mLevel = LevelFactory.loadExternal(App.options.getExternal());
			else
				mLevel = LevelFactory.loadInternal(name);
			mLevelLayout = mLevel.getLayout();
		} catch (JAXBException e) {
			System.err.println("JAXBException while loading level "+name+": " + e.getMessage());
			e.printStackTrace(System.err);
			System.exit(1);
		} catch (IOException e) {
			System.err.println("IOException:  while loading level "+name+"" + e.getMessage());
			e.printStackTrace(System.err);
			System.exit(1);
		}
		mLevel.setGameWorld(gameWorld);
		mLevelObject = new LevelObject(mLevelLayout);
		
		mPlayer = new PlayerObject(mLevel.getSpawnPositionX(),
				mLevel.getSpawnPositionY());
		root.setPlayer(mPlayer);
		mGameObjectLookUpMap.put("player", mPlayer);
		mCamera = new Camera(mPlayer);

		cWorld = new CollisionWorld();
		List<GameObject> newLevelObjects = new ArrayList<GameObject>();
		cWorld.add(mLevelObject.getStaticObjects());
		for (SimpleObject mo : mLevel.getObjects()) {
			newLevelObjects.add(mo);
			cWorld.add(mo);
			if (mo.getName() != null)
				mGameObjectLookUpMap.put(mo.getName(), mo);
			mo.setGameWorld(gameWorld);
		}
		newLevelObjects.add(mPlayer);
		mPlayer.setGameWorld(gameWorld);
		cWorld.add(mPlayer);

		mLevel.doWelcomeMessage();
		App.globals.setCurrentLevel(mLevel, mLevel.getLevelName());
		
		gameWorld.updateTests(mLevel);
		loadNextLevel = false;
		mGameObjects = Collections.synchronizedList(newLevelObjects);
		App.globals.save();
	}

	/**
	 * Get game objects Currently used by Game Renderer
	 * 
	 * @return Array of the game objects
	 */
	public List<GameObject> getSortedGameObjects() {
		renderObjects.clear();
		if (mCamera == null) return renderObjects;
		view.setFrame(-mCamera.getDispX(), -mCamera.getDispY(), ResourceFactory.SCREEN_WIDTH, ResourceFactory.SCREEN_HEIGHT);
		
		for (GameObject go : mGameObjects) {
			if (go instanceof CompositeObject) {
				addCompositeToRenderObjects(go);
			} else
				if (view.intersects(new Rectangle(go.getX(), go.getY(), go.getHeight(), go.getWidth())));
					renderObjects.add(go);
		}
		
		renderObjects.addAll(mLevelObject.getDrawingTiles(view));
		Collections.sort(renderObjects, mGameObjectComparator);
		return renderObjects;
	}
	
	private void addCompositeToRenderObjects(GameObject object) {
		for (SimpleObject go : ((CompositeObject) object).getCells())
			if (go instanceof CompositeObject) {
				addCompositeToRenderObjects(go);
			} else
				renderObjects.add(go);
	}

	/**
	 * Get current level. Used by GameWindow for rendering
	 * 
	 * @return a LevelObject
	 */
	public LevelObject getLevel() {
		return mLevelObject;
	}

	/**
	 * Adds a game object to the game universes (rendering and collision
	 * universes)
	 * 
	 * @param o
	 *            - the GameObject
	 */
	public void addGameObject(GameObject o) {
		mGameObjects.add(o);
		cWorld.add(o);
		if (o.getName() != null)
			mGameObjectLookUpMap.put(o.getName(), o);
		o.setGameWorld(gameWorld);
	}

	/**
	 * Removes a object from the game universes (rendering and collision
	 * universes)
	 * 
	 * @param o
	 */
	public void removeGameObject(GameObject o) {
		mGameObjects.remove(o);
		cWorld.remove(o);
		if (o.getName() != null)
			mGameObjectLookUpMap.remove(o);
		o.setGameWorld(null);
	}

	public void showMessageBox(String text) {
		MessageBox msg = new MessageBox(root, text);
		msg.openPopupCentered(gui.getWidth() * 1 / 2, gui.getHeight() * 1 / 2);
	}

	/**
	 * Main game thread
	 */
	long lastFrameTime = System.nanoTime();
	long currentFrameTime, delta = 0;
	long[] benchmark = new long[10];
	long tmp = 0;
	long fps, frameTime = 0;
	int frameCounter = 0;
	double deltams;
	
	private Runnable transitionAction;
	private boolean transitionActionRan = false;
	private SimpleObject tempGameObject = null;
	private void runGame() {
		if (loadNextLevel == true || (mPlayer != null && mPlayer.isDead())) {
			next_state = GAME_STATE_IN_GAME;
			state = GAME_STATE_TRANSITION;
			transitionAction = new Runnable() {
				@Override
				public void run() {
					root.showHUD();
					loadNextLevel(nextLevel);
					loadNextLevel = false;
				}
			};
			return;
		}
		
		if (!root.statusLine.hasText()) {
			tempGameObject = getClosestObject(50);
			if (tempGameObject != null && tempGameObject.getCloseMessage() != null)
				root.statusLine.setText(tempGameObject.getCloseMessage());
		}
		

		tmp = System.nanoTime();
		if (root.gameHasFocus()) {
			if (Keyboard.isKeyDown(Keyboard.KEY_RIGHT))
				mPlayer.setMoveRight(true);
			else
				mPlayer.setMoveRight(false);
			if (Keyboard.isKeyDown(Keyboard.KEY_LEFT))
				mPlayer.setMoveLeft(true);
			else
				mPlayer.setMoveLeft(false);
			if (Keyboard.isKeyDown(Keyboard.KEY_DOWN))
				mPlayer.setMoveDown(true);
			else
				mPlayer.setMoveDown(false);
			if (Keyboard.isKeyDown(Keyboard.KEY_UP))
				mPlayer.setMoveUp(true);
			else
				mPlayer.setMoveUp(false);
		} else {
			mPlayer.setMoveDown(false);
			mPlayer.setMoveUp(false);
			mPlayer.setMoveRight(false);
			mPlayer.setMoveLeft(false);
		}
		benchmark[1] += System.nanoTime() - tmp;

		tmp = System.nanoTime();

		
		for (GameObject o : mGameObjects)
			o.move(deltams);
		for (GameObject o : mLevelObject.getStaticObjects())
			o.move(deltams);

		cWorld.compute();
		benchmark[2] += System.nanoTime() - tmp;

		tmp = System.nanoTime();
		benchmark[3] += System.nanoTime() - tmp;
		if (ResourceFactory.CALC_PERFORMANCE_LOG_ENABLED
				&& frameCounter % 1000 == 0) {
			fps = frameCounter * 1000000000L / frameTime;
			System.out.println("FPS: " + fps);
			System.out.println("Total frame time: \t\t" + benchmark[0]);
			System.out.println("Time in input read: \t\t" + benchmark[1]);
			System.out.println("Time in move and collide: \t" + benchmark[2]);
			System.out
					.println("Time in set camera position: \t" + benchmark[3]);
			System.out.println("Last delta: " + delta);
			Arrays.fill(benchmark, 0);
		}

		
	}
	
	private void runMainMenu() {
		if (!mainMenu.isOpen()) {
			root.hideAll();
			mainMenu.openPopupCentered();
		}
		
	}
	
	public void startNewGame() {
		loadLevel(App.globals.getFirstLevelName());
		state = GAME_STATE_IN_GAME;
	}
	
	private void runTransition() {
		if (!transitionActionRan) {
			transition_progress += deltams;
			if (transition_progress >= 1.0f) {
				if (transitionAction != null) transitionAction.run();
				transitionActionRan = true;
			}
		} else {
			transition_progress -= deltams;
			if (transition_progress <= 0) {
				transition_progress = 0;
				transitionActionRan = false;
				transitionAction = null;
				state = next_state;
			}
		}
	}
	
	private void runGameMenu() {
		
	}

	public void run() {
		frameCounter++;
		currentFrameTime = System.nanoTime();

		benchmark[0] += delta;
		deltams = (delta > 100000000) ? 0.1 : 1.0 * delta / 1000000000;
		
		switch (state) {
		case GAME_STATE_INTRO:
			// TODO: we do not have any intro
			state = GAME_STATE_MAIN_MENU;
			break;
		case GAME_STATE_MAIN_MENU:
			runMainMenu();
			break;
		case GAME_STATE_TRANSITION:
			runTransition();
			break;
		case GAME_STATE_IN_GAME:
			runGame();
			break;
		case GAME_STATE_GAME_MENU:
			if (!gameMenu.isOpen())
				state = GAME_STATE_IN_GAME;
			break;
		case GAME_STATE_GAME_OVER:
			break;
		case GAME_STATE_GAME_END:
			break;
		default:
			break;
		}
		
		delta = currentFrameTime - lastFrameTime;
		frameTime += delta;
		lastFrameTime = currentFrameTime;
	}

	/**
	 * Get current camera's position
	 * 
	 * @return a point, unless we do not have a camera, in which case, we are
	 *         screwed
	 */
	public Point2D.Double getCameraPosition() {
		if (mCamera == null)
			return new Point2D.Double();
		return new Point2D.Double(mCamera.getDispX(), mCamera.getDispY());
	}

	/**
	 * Finds the GameObject that is closest to the player. Returns null if no
	 * objects are within threshold
	 * 
	 * @return
	 */
	public SimpleObject getClosestObject(double threshold) {
		GameObject ret = null;
		for (GameObject o : mGameObjects) {
			if (!(o instanceof SimpleObject)) continue;
			if (o.isDead()) continue;
			double distance = o.distanceTo(mPlayer);
			if (distance < threshold) {
				ret = o;
				threshold = distance;
			}
		}
		return (SimpleObject)ret;
	}

	/**
	 * I really do not know what was I thinking about when I ate that stake, but
	 * I know now that this should probably be moved to the window
	 * 
	 * @return
	 */
	public GUI getGUI() {
		return gui;
	}

	/**
	 * The win... probably
	 */
	public void windowReady() {
		try {
			LWJGLRenderer lwjglRenderer = new LWJGLRenderer();
			lwjglRenderer.syncViewportSize();
			theme = ThemeManager.createThemeManager(
					Game.class.getResource("/twl/simple.xml"), lwjglRenderer);
			// theme =
			// ThemeManager.createThemeManager(Game.class.getResource("/twl_test/pythia.xml"),
			// lwjglRenderer);
			gui = new GUI(root, lwjglRenderer);
			gui.applyTheme(theme);
			gui.setSize();
			root.helpPanel.setSize(gui.getWidth(),
					root.helpPanel.getPreferredHeight());
			root.statusLine.setSize(gui.getWidth(),
					root.statusLine.getPreferredHeight());
			root.statusLine.setPosition(0,
					gui.getHeight() - root.statusLine.getPreferredHeight());
			root.statusLine.setText("Welcome!");

			root.consoleDialog.setSize(gui.getWidth(), gui.getHeight() * 1 / 3);
			root.consoleDialog.setPosition(0, root.helpPanel.getHeight());

			root.codeEditorDialog.setSize(gui.getWidth() * 1 / 2,
					gui.getHeight() - root.statusLine.getHeight()
							- root.helpPanel.getHeight());
			root.codeEditorDialog.setPosition(0, root.helpPanel.getHeight());
			mainMenu = new MainMenu(root, this);
			gameMenu = new GameMenu(root, this);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * Over here we can handle single key presses and releases, which were not
	 * consumed by the all-mighty, all-hungry TWL
	 */
	@Override
	public boolean handleEvent(Event evt) {
		switch (evt.getType()) {
		case KEY_PRESSED:
			switch (evt.getKeyCode()) {
			case Event.KEY_RETURN:
				GameObject go = getClosestObject(50);
				if (go != null) {
					if (go instanceof SimpleObject) {
						SimpleObject so = (SimpleObject) go;
						so.talkTo();
					}
				}
				return true;
			case Event.KEY_F9:
				Tester.getTester().showRemainingObjectives();
				return true;
			case Event.KEY_F10:
				gameMenu.openPopupCentered();
				state = GAME_STATE_GAME_MENU;
				return true;
			case Event.KEY_X:
				if (App.options.isDebug() && mPlayer != null) {
					System.out.printf("Player's coordinates:\n\tX: %d\n\tY: %d\n", mPlayer.getX(), mPlayer.getY());
					return true;
				}
				break;
			default:
				break;
			}
			break;

		case KEY_RELEASED:
			break;
		}
		return false;
	}

	public void showMessage(String text, long ms, Position o, int dispY) {
		if (root != null)
			root.showDialog(text, ms, mCamera, o, dispY);
	}

	public void statusUPdate(String text) {
		if (root != null)
			root.statusLine.setText(text);
	}

	class GameObjectComparator implements Comparator<GameObject> {
		@Override
		public int compare(GameObject o1, GameObject o2) {
			if (o1 == null)
				if (o2 == null)
					return 0;
				else
					return -1;
			if (o2 == null)
				return 1;

			if (o1.getZindex() != o2.getZindex())
				return o1.getZindex() - o2.getZindex();
			return (o1.getY() + o1.getHeight() / 2)
					- (o2.getY() + o2.getHeight() / 2);
		}
	}

	public Map<String, GameObject> getGameObjectMap() {
		return mGameObjectLookUpMap;
	}

	public void showConsole(String text) {
		if (root != null) root.showConsole(text);
	}
	
	public void showConsole() {
		if (root != null) root.showConsole();
	}

	public void showEditor() {
		if (root != null) root.showEditor();
	}

	public GameObject getObjectByName(String name) {
		return mGameObjectLookUpMap.get(name);
	}

	public CodeEditorDialog getEditor() {
		return root.codeEditorDialog;
	}

	public Camera getCamera() {
		return mCamera;
	}
	
	public int getState() {
		return state;
	}

	public void restartLevel() {
		loadNextLevel = true;
	}

	public void quitGame() {
		transitionAction = new Runnable() {
			public void run() {
				mCamera = null;
				root.hideAll();
				mPlayer = null;
				mLevel = null;
				mLevelObject = null;
				mGameObjects.clear();
				App.globals.getParser().clear();
				mGameObjectLookUpMap.clear();		
			}
		};
		next_state = GAME_STATE_MAIN_MENU;
		state = GAME_STATE_TRANSITION;
		
		
	}

}
