package org.ntnu.pythia.engine;

import java.awt.Color;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

import org.lwjgl.opengl.Display;

import org.ntnu.pythia.engine.components.CodeEditorDialog;
import org.ntnu.pythia.engine.lwjgl.LWJGLColorBox;
import org.ntnu.pythia.engine.lwjgl.LWJGLGameWindow;
import org.ntnu.pythia.engine.lwjgl.LWJGLSoundObject;
import org.ntnu.pythia.engine.lwjgl.LWJGLSprite;
import org.ntnu.pythia.game.AnimationObject;
import org.ntnu.pythia.game.AnimationTemplate;
import org.ntnu.pythia.game.RenderObject;
import org.ntnu.pythia.game.SoundObject;

public class ResourceFactory {
	public static final int SCREEN_WIDTH = 1024;
	public static final int SCREEN_HEIGHT = 768;
	
	public static final String WINDOW_TITLE = "Pythia";
	public static final boolean CALC_PERFORMANCE_LOG_ENABLED = false;
	public static final boolean RENDER_PERFORMANCE_LOG_ENABLED = false;
	public static boolean RENDER_COLLISION_MODELS = false;
	
	public static final int FPS_LIMIT = 60;
	
	private static GameWindow gameWindow;
	private static GameWorld gameWorld;
	private static Game game;
	
	private static final Map<String, RenderObject> spriteStore = new ConcurrentHashMap<String, RenderObject>();
	private static final Map<Color, RenderObject> colorBoxStore  = new ConcurrentHashMap<Color, RenderObject>();
	private static final Map<String, AnimationTemplate> animationStore = new ConcurrentHashMap<String, AnimationTemplate>();
	private static final Map<String, SoundObject> soundStore = new ConcurrentHashMap<String, SoundObject>();
	
	public static final String CURRENT_FOLDER = System.getProperty("user.dir");
	
	private ResourceFactory() {}
	
	/**
	 * Method for getting a game window with chosen rendering method
	 * @return Returns an instance of a game window
	 */
	public static GameWindow getGameWindow() {
		DebugLog.println("Graphics card: " + Display.getAdapter() + ", version: " + Display.getVersion());
		gameWindow = new LWJGLGameWindow(WINDOW_TITLE, SCREEN_WIDTH, SCREEN_HEIGHT);
		return gameWindow;
	}
	
	public static SoundObject getSoundObject(String ref, String filename) {
		SoundObject result = new LWJGLSoundObject(filename);
		soundStore.put(ref, result);
		return result;
	}
	
	public static SoundObject getSoundObject(String ref) {
		return soundStore.get(ref);
	}

	/**
	 * Get a sprite based on color (sprite size is 20x20)
	 * @param color The color of the sprite
	 * @return An instance of a sprite in a current rendering model
	 */
	public static RenderObject getRenderObject(Color color) {
		RenderObject result = colorBoxStore.get(color);
		if (result == null) {
			result = new LWJGLColorBox(color);
			colorBoxStore.put(color, result);
		}
		return result;
	}
	
	public static RenderObject getRenderObject(String ref, String filename) {
		RenderObject result = new LWJGLSprite((LWJGLGameWindow)gameWindow, filename);
		spriteStore.put(ref, result);
		return result;
	}
	
	public static RenderObject getRenderObject(String ref) {
		if (animationStore.containsKey(ref))
			return animationStore.get(ref).getAnimation();
		return spriteStore.get(ref);
	}
	
	public static void unloadRenderObjects() {
		spriteStore.clear();
	}
	
	public static AnimationObject getAnimationObject(String ref, String[] refs, String[] filenames, double length) {
		AnimationTemplate template = new AnimationTemplate(refs, filenames, length);
		animationStore.put(ref, template);
		return template.getAnimation();
	}
	
	public static AnimationObject getAnimationObject(String ref) {
		if (animationStore.containsKey(ref))
			return animationStore.get(ref).getAnimation();
		System.err.println("No animation '" + ref + "' found!");
		return null;
	}
	
	public static String getTextString(String ref) {
		StringBuilder sb =  new StringBuilder();
		try{
			InputStream is;
			if(App.options.isExternal())
				is = new FileInputStream(new File(App.options.getExternal(), ref));
			else
			is = ResourceFactory.class.getResourceAsStream(ref);
			if(is == null)
				throw new RuntimeException("Error while reading textwindow content: '" + ref + "'.");
			BufferedReader br = new BufferedReader(new InputStreamReader(is));
			while(br.ready())
				sb.append(br.readLine());
			br.close();
		}catch (IOException e) {
			throw new RuntimeException("Error while reading textwindow content: '" + ref + "'.", e);
		}
		return sb.toString();
	}
	
	public static Game getGame() {
		if(game == null)
			game = new Game();
		return game;
	}
	
	public static GameWorld getGameWorld(){
		if(gameWorld == null){
			if(game == null)
				game = getGame();
			gameWorld = new GameWorld(game);
		}
		return gameWorld;
				
	}
	
	public static void setRenderCollision(boolean value) {
		RENDER_COLLISION_MODELS = value;
	}

	public static CodeEditorDialog getEditor() {
		return game.getEditor();
	}
}
