package org.ntnu.pythia.engine.components;

import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.ntnu.pythia.engine.App;
import org.ntnu.pythia.python.ListenableVariableMap;
import org.ntnu.pythia.python.ListeningEditor;
import org.ntnu.pythia.python.UserScript;
import org.python.core.PyException;
import org.python.core.PyFrame;
import org.python.core.PyList;
import org.python.core.PyObject;
import org.python.core.PyStringMap;

import de.matthiasmann.twl.Button;
import de.matthiasmann.twl.DialogLayout;
import de.matthiasmann.twl.EditField;
import de.matthiasmann.twl.Event;
import de.matthiasmann.twl.Label;
import de.matthiasmann.twl.ResizableFrame;
import de.matthiasmann.twl.ScrollPane;
import de.matthiasmann.twl.Widget;

public class CodeEditorDialog extends ResizableFrame implements ListeningEditor {
	public static final int SCRIPT_STATE_READY = 0;
	public static final int SCRIPT_STATE_STEPPING = 4;
	public static final int SCRIPT_STATE_EXECUTING = 1;
	public static final int SCRIPT_STATE_HALTED = 2;
	
	private int mState = SCRIPT_STATE_READY;
	
	private final CodeEditField codeEditField;
	private final LineNumbersField lineNumbers;
	private final ScrollPane scrollPane;
	
	private final EditField editStatus;
	
	private final EditField consoleOutput;
	private UserScript script;
	private Thread scriptThread;
	
	private final ConsoleOutputStream consoleOutputStream;
	
	private List<Integer> breakpointLines = new ArrayList<Integer>();
	
	class CodeEditScrollPane extends ScrollPane {
		public CodeEditScrollPane(Widget content) {
			super(content);
		}
		@Override
		public int getPreferredWidth() {
			if (getParent() == null) return 0;
			if (lineNumbers == null) return super.getPreferredWidth();
			return getParent().getWidth() - lineNumbers.getWidth();
		}
		@Override
		public int getPreferredHeight() {
			if (getParent() == null) return 0;
			return getParent().getHeight();
		}
	}
	
	class CustomDialogLayout extends DialogLayout {
		@Override
		public int getPreferredHeight() {
			if (getParent() == null) return 0;
			return getParent().getHeight();
		}
	}
	
	public CodeEditorDialog() {
		codeEditField = new CodeEditField();
		codeEditField.setText("# write code here");
		codeEditField.setTheme("/codeeditfield");
		lineNumbers = new LineNumbersField(codeEditField);
		lineNumbers.setTheme("/linenumberfield");
		scrollPane = new CodeEditScrollPane(codeEditField);
		scrollPane.setTheme("/scrollpane");
		scrollPane.setExpandContentSize(true);
		
		DialogLayout codeEditorLayout = new CustomDialogLayout();
		codeEditorLayout.setTheme("/dialoglayout");
		codeEditorLayout.setVerticalGroup(codeEditorLayout.createParallelGroup(lineNumbers, scrollPane));
		codeEditorLayout.setHorizontalGroup(codeEditorLayout.createSequentialGroup(lineNumbers).addGap().addGroup(codeEditorLayout.createSequentialGroup(scrollPane)));
		
		// Status Field
		Label lblStatus = new Label("Status: ");
		lblStatus.setTheme("/label");
		editStatus = new EditField();
		editStatus.setTheme("/editfield");
		editStatus.setReadOnly(true);
		editStatus.setText("Ready!");
		
		// Output console
		consoleOutput = new EditField();
		consoleOutput.setTheme("/editfield_multiline");
		consoleOutput.setReadOnly(true);
		consoleOutput.setMultiLine(true);
		consoleOutput.setMaxTextLength(Integer.MAX_VALUE);
		CustomScrollPane consoleScrollPane = new CustomScrollPane(consoleOutput); 
		consoleScrollPane.setTheme("/scrollpane");
		consoleScrollPane.setExpandContentSize(true);
		consoleOutputStream = new ConsoleOutputStream(consoleOutput);
		
		
		Button btnRun = new Button("Run");
		Button btnStep = new Button("Step");
		btnRun.setTheme("/button");
		btnStep.setTheme("/button");
		
		DialogLayout codeDialogLayout = new DialogLayout();
		codeDialogLayout.setTheme("/dialoglayout");
		codeDialogLayout.setHorizontalGroup(
				codeDialogLayout.createParallelGroup(codeEditorLayout)
					.addGroup(codeDialogLayout.createSequentialGroup(lblStatus, editStatus))
					.addGroup(codeDialogLayout.createSequentialGroup(consoleScrollPane)));
		codeDialogLayout.setVerticalGroup(
				codeDialogLayout.createSequentialGroup(codeEditorLayout)
					.addGap()
					.addGroup(codeDialogLayout.createSequentialGroup()
							.addGroup(codeDialogLayout.createParallelGroup(lblStatus, editStatus))
							.addGroup(codeDialogLayout.createParallelGroup(consoleScrollPane))));
				
		setTheme("resizableframe-title");
		setTitle("Editor");
		add(codeDialogLayout);
	}
	
	@Override
	public void setVisible(boolean value) {
		super.setVisible(value);
		codeEditField.requestKeyboardFocus();
	}
	
	private void setState(int newState) {
		mState = newState;
		switch (mState) {
		case SCRIPT_STATE_READY:
			editStatus.setText("Ready!");
			break;
		case SCRIPT_STATE_EXECUTING:
		case SCRIPT_STATE_STEPPING:
			editStatus.setText("Running!");
			break;
		case SCRIPT_STATE_HALTED:
			editStatus.setText("Paused!");
			break;
		default:
			editStatus.setText("Unknown state error!");
			break;
		}
	}
	
	@Override
	public boolean handleEvent(Event evt) {
		if (evt.getType() == Event.Type.KEY_PRESSED) {
			switch (evt.getKeyCode()) {
			case Event.KEY_TAB:
				return codeEditField.handleEvent(evt);
			case Event.KEY_F5:
				if (scriptThread.isAlive())
					break;
				consoleOutputStream.clear();
				setState(SCRIPT_STATE_EXECUTING);
				App.output.addStream(consoleOutputStream);
				breakpointLines = lineNumbers.getBreakpoints();
				codeEditField.setReadOnly(true);
				scriptThread = new Thread(script, "LoserScript");
				scriptThread.start();
				return true;
			case Event.KEY_F6:
				if (mState == SCRIPT_STATE_HALTED) {
					breakpointLines = lineNumbers.getBreakpoints();
					setState(SCRIPT_STATE_STEPPING);
					script.step();
					return true;
				}
			case Event.KEY_F7:
				if (mState == SCRIPT_STATE_HALTED) {
					breakpointLines = lineNumbers.getBreakpoints();
					setState(SCRIPT_STATE_EXECUTING);
					script.unhalt();
					return true;
				}
			case Event.KEY_F8:
				if (mState == SCRIPT_STATE_HALTED || mState == SCRIPT_STATE_EXECUTING || mState == SCRIPT_STATE_STEPPING) {
					script.stop();
					return true;
				}
			case Event.KEY_F12:
				if (mState == SCRIPT_STATE_HALTED) {
					PyFrame pf = script.getFrame();
					PyList vars = ((PyStringMap)pf.f_locals).items();
					Iterator<PyObject> it = vars.iterator();
					while(it.hasNext()) {
						PyObject o = it.next();
						if (o.isDataDescr())
							System.out.println("key: " + o.__str__() + " value: " + o.asString());
					}
					System.out.println("done");
					return true;
				}
			default:
				break;
			}
		}
		return super.handleEvent(evt);
	}

	@Override
	public void error(String message, int lineNumber, PyException e) {
		codeEditField.setExceptionLine(lineNumber);
		PrintStream ps = new PrintStream(consoleOutputStream);
		e.printStackTrace(ps);
		ps.flush();

	}

	@Override
	public void executing(int lineNumber) {
		Integer i = new Integer(lineNumber);
		if (mState == SCRIPT_STATE_STEPPING) {
			codeEditField.setExceptionLine(lineNumber);
			script.halt();
			setState(SCRIPT_STATE_HALTED);
		}
		if (breakpointLines.contains(i)) {
			codeEditField.setExceptionLine(lineNumber);
			script.halt();
			setState(SCRIPT_STATE_HALTED);
		}
	}

	@Override
	public String getProgram() {
		return codeEditField.getText();
	}

	@Override
	public void finished() {
		setState(SCRIPT_STATE_READY);
		App.output.removeStream(consoleOutputStream);
		codeEditField.setReadOnly(false);
		codeEditField.setExceptionLine(CodeEditField.NO_EXCEPTION);
	}
	
	@Override
	public void destroy() {
		super.destroy();
		if (scriptThread.isAlive()) script.stop();
	}
	
	public void levelLoaded(){
		script = new UserScript(App.globals.getParser(), this);
		scriptThread = new Thread(script, "LoserScript");
	}

	@Override
	public void setProgram(String text) {
		codeEditField.setText(text);
	}

}
