package org.ntnu.pythia.engine.components;

import java.io.IOException;
import java.io.OutputStream;
import java.util.LinkedList;
import java.util.List;

import org.ntnu.pythia.engine.App;
import org.ntnu.pythia.python.PythiaParser;
import org.python.core.Py;
import org.python.core.PyException;

import de.matthiasmann.twl.DialogLayout;
import de.matthiasmann.twl.EditField;
import de.matthiasmann.twl.Event;
import de.matthiasmann.twl.GUI;
import de.matthiasmann.twl.ResizableFrame;
import de.matthiasmann.twl.ScrollPane;
import de.matthiasmann.twl.TextArea;
import de.matthiasmann.twl.TextWidget;
import de.matthiasmann.twl.Widget;
import de.matthiasmann.twl.model.EditFieldModel;
import de.matthiasmann.twl.model.StringModel;
import de.matthiasmann.twl.renderer.Font;
import de.matthiasmann.twl.textarea.HTMLTextAreaModel;

public class ConsoleDialog extends ResizableFrame {
//	protected final HTMLTextAreaModel tam;
	protected final EditField editField;
	protected final CustomScrollPane scrollPane;
//	protected final TextArea textArea;
	private final EditField consoleOutput;
	protected String content = 
	" ____  __    __  _____   _    _    _      __\n"+
	"|    \\ \\ \\  / / |_   _| | |  | |  | |    /  \\ \n"+
	"| () |  \\ \\/ /    | |   | |__| |  | |   / /\\ \\ \n"+
	"| ___/   \\  /     | |   |  __  |  | |  / ____ \\ \n"+
	"| |      / /      | |   | |  | |  | | / /    \\ \\ \n"+
	"|_|     /_/       |_|   |_| Pythia Console 0.2.0\n";

	private List<String> history = new LinkedList<String>();
	private int indentation;
	private int returnTo;
	private int historyCounter = 0;

	private final PythiaParser parser = App.globals.getParser();
	private ConsoleOutputStream consoleOut;

	public ConsoleDialog() {
		consoleOutput = new EditField();
		consoleOutput.setTheme("/editfield_multiline");
		consoleOutput.setReadOnly(true);
		consoleOutput.setMultiLine(true);
//		consoleOutput.setText(content);
		consoleOut = new ConsoleOutputStream(consoleOutput);
		App.output.addStream(consoleOut);
		try {
			consoleOut.write(content.getBytes());
		} catch (IOException e) {
			e.printStackTrace();
		}

		scrollPane = new CustomScrollPane(consoleOutput);
		scrollPane.setTheme("/scrollpane");
		scrollPane.setFixed(ScrollPane.Fixed.HORIZONTAL);
		scrollPane.setExpandContentSize(true);

		editField = new ConsoleEditField();
		editField.setTheme("/editfield");
		editField.setMultiLine(false);
		editField.setText("");

		DialogLayout box = new DialogLayout();
		box.setTheme("/optionsdialog");
		box.setHorizontalGroup(box.createParallelGroup().addWidget(scrollPane).addWidget(editField));
		box.setVerticalGroup(box.createSequentialGroup().addWidgets(scrollPane, editField));

		setTheme("resizableframe-title");
		add(box);
		setTitle("Console");
	}
	
	public void output(String msg) {
//		msg = msg.replaceAll("\t", "&#09;");//XXX:generell escaping kanskje?
		try {
			consoleOut.write((msg + "\n").getBytes());
		} catch (IOException e) {
			e.printStackTrace();
		}
		
		scrollPane.setToScroll();
	}


	public void execute() {
		String line = editField.getText();
		if(!line.trim().equals(""))
			output("$- " + line);
		history.add(line);
		historyCounter = history.size();

		if(indentation > 0){
			int tabs = countTabs(line);
			if(tabs == 0){
				pushToPython();//feed buffer
				indentation = 0;//reset indent-count
			}else if(indentation > tabs)//returned from inner definition, but not all the way to the bottom.
				indentation = tabs;
		}
		if((line.charAt(line.length() -1) == ':')){//start of definition
			if(indentation == 0)
				returnTo = history.size() -1; //flush everything back to this line.
			indentation++;
		}
		if(indentation == 0){
			//single line
			returnTo = history.size() -1;
			pushToPython();
		}

		//place the correct number of tabs on the start of the next input line
		editField.setText(createTabs());
	}
	
	private void scriptDone() {
		editField.setReadOnly(false);
	}

	private void pushToPython(){
		StringBuilder builder = new StringBuilder();
		for(int i = returnTo; i < history.size(); i++)
			builder.append(history.get(i)+"\n");
		editField.setReadOnly(true);
		
			final String script = builder.toString();
			new Thread(new Runnable() {
				@Override
				public void run() {
					try {
						parser.exec(script, true);
					} catch (PyException e) {
						String out = e.value.toString();
						output(out);//TODO: red text for errors?
						if (!e.match(Py.KeyboardInterrupt))
							e.printStackTrace();
					} catch (Throwable e) {
						output("*** Interrupted! ***");
						//System.out.println(e);
						System.out.println(e.getClass());
					} finally {
						scriptDone();
					}
				}
			}).start();
	}

	private static int countTabs(String in){
		int count = 0;
		for (int i = 0; i < in.length(); i++) {
			if(in.charAt(i) == '\t')
				count++;
			else
				return count;
		}
		return count;
	}

	@Override
	public void setVisible(boolean value) {
		super.setVisible(value);
		editField.requestKeyboardFocus();
	}

	@Override
	public boolean handleEvent(Event evt) {
		if (evt.getType() == Event.Type.KEY_PRESSED && evt.getKeyCode() == Event.KEY_TAB) {
			return editField.handleEvent(evt);
		}
		return super.handleEvent(evt);
	}

	private String createTabs(){
		StringBuilder sb = new StringBuilder();
		for (int i = 0; i < indentation; i++) 
			sb.append('\t');
		return sb.toString();
	}

	class ConsoleEditField extends EditField {
		@Override
		public boolean handleEvent(Event evt) {
			if (isReadOnly() && evt.getType() == Event.Type.KEY_PRESSED && 
					((evt.getModifiers() & Event.MODIFIER_CTRL) != 0) && evt.getKeyCode() == Event.KEY_Z) {
					parser.interruptThread();
					return true;
				}
			if (evt.getType() == Event.Type.KEY_PRESSED && evt.getKeyCode() == Event.KEY_TAB) {
				insertText("\t");
				return true;
			}
			if (evt.getType() == Event.Type.KEY_PRESSED && evt.getKeyCode() == Event.KEY_RETURN ) {
				execute();
				return true;
			}
			if (evt.getType() == Event.Type.KEY_PRESSED && evt.getKeyCode() == Event.KEY_UP) {
				if(!history.isEmpty() && !editField.isReadOnly()){
					historyCounter--;
					if(historyCounter < 0)
						historyCounter = 0;
					setText(history.get(historyCounter));
					return true;
				}
			}
			if (evt.getType() == Event.Type.KEY_PRESSED && evt.getKeyCode() == Event.KEY_DOWN) {
				if(!history.isEmpty() && !editField.isReadOnly()){
					historyCounter++;
					if(historyCounter > history.size()-1){
						historyCounter = history.size();
						setText(createTabs());
					}else
						setText(history.get(historyCounter));
					return true;
				}
			}
			return super.handleEvent(evt);
		}
	}

	public void showLogo() {
		try {
			consoleOut.flush();
			consoleOut.write(content.getBytes());
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

}
