package org.ntnu.pythia.engine.components;

import java.util.ArrayList;
import java.util.List;

import org.ntnu.pythia.engine.components.CodeEditField.LineCountListener;

import de.matthiasmann.twl.AnimationState;
import de.matthiasmann.twl.Event;
import de.matthiasmann.twl.GUI;
import de.matthiasmann.twl.InfoWindow;
import de.matthiasmann.twl.Label;
import de.matthiasmann.twl.TextWidget;
import de.matthiasmann.twl.ThemeInfo;
import de.matthiasmann.twl.Widget;
import de.matthiasmann.twl.model.DefaultEditFieldModel;
import de.matthiasmann.twl.model.EditFieldModel;
import de.matthiasmann.twl.renderer.Font;
import de.matthiasmann.twl.renderer.Image;
import de.matthiasmann.twl.renderer.AnimationState.StateKey;
import de.matthiasmann.twl.utils.CallbackSupport;
import de.matthiasmann.twl.utils.TextUtil;

public class LineNumbersField extends Widget implements LineCountListener {
    public static final StateKey STATE_ERROR = StateKey.get("error");
    public static final StateKey STATE_READONLY = StateKey.get("readonly");
    public static final StateKey STATE_HOVER = StateKey.get("hover");

    public interface Callback {
        /**
         * Gets called for any change in the edit field, or when ESCAPE or RETURN was pressed
         *
         * @param key One of KEY_NONE, KEY_ESCAPE, KEY_RETURN, KEY_DELETE
         * @see Event#KEY_NONE
         * @see Event#KEY_ESCAPE
         * @see Event#KEY_RETURN
         * @see Event#KEY_DELETE
         */
        public void callback(int key);
    }

    final EditFieldModel editBuffer;
    private final TextRenderer textRenderer;

    int numberOfLines;
    boolean pendingScrollToCursor;
    boolean pendingScrollToCursorForce;
    private int maxTextLength = Short.MAX_VALUE;

    private int columns = 5;
    Image selectionImage;
    private Object errorMsg;
    private Callback[] callbacks;
    private boolean textLongerThenWidget;
    private boolean forwardUnhandledKeysToCallback;

    private InfoWindow errorInfoWindow;
    private Label errorInfoLabel;

    /**
     * Creates a new EditField with an optional parent animation state.
     *
     * Unlike other widgets which use the passed animation state directly,
     * the EditField always creates it's animation state with the passed
     * one as parent.
     *
     * @param parentAnimationState
     * @param editFieldModel the edit field model to use
     * @see AnimationState#AnimationState(de.matthiasmann.twl.AnimationState) 
     */
    public LineNumbersField(CodeEditField editor, AnimationState parentAnimationState, EditFieldModel editFieldModel) {
        super(parentAnimationState, true);

        if(editFieldModel == null) {
            throw new NullPointerException("editFieldModel");
        }

        this.editBuffer = editFieldModel;
        this.textRenderer = new TextRenderer(getAnimationState());
        
        textRenderer.setTheme("renderer");
        textRenderer.setClip(true);
        
        add(textRenderer);
        setCanAcceptKeyboardFocus(false);
        setDepthFocusTraversal(false);
        
        setLineCount(linecount, scrollY);
        
        if (editor != null) editor.setLineCountListener(this);
    }

    /**
     * Creates a new EditField with an optional parent animation state.
     *
     * Unlike other widgets which use the passed animation state directly,
     * the EditField always creates it's animation state with the passed
     * one as parent.
     *
     * @param parentAnimationState
     * @see AnimationState#AnimationState(de.matthiasmann.twl.AnimationState)
     */
    public LineNumbersField(CodeEditField editor, AnimationState parentAnimationState) {
        this(editor, parentAnimationState, new DefaultEditFieldModel());
    }

    public LineNumbersField(CodeEditField editor) {
        this(editor, null);
    }

    public void addCallback(Callback cb) {
        callbacks = CallbackSupport.addCallbackToList(callbacks, cb, Callback.class);
    }

    public void removeCallback(Callback cb) {
        callbacks = CallbackSupport.removeCallbackFromList(callbacks, cb);
    }

    public boolean isForwardUnhandledKeysToCallback() {
        return forwardUnhandledKeysToCallback;
    }

    /**
     * Controls if unhandled key presses are forwarded to the callback or not.
     * Default is false. If set to true then the EditField will consume all key
     * presses.
     *
     * @param forwardUnhandledKeysToCallback true if unhandled keys should be forwarded to the callbacks
     */
    public void setForwardUnhandledKeysToCallback(boolean forwardUnhandledKeysToCallback) {
        this.forwardUnhandledKeysToCallback = forwardUnhandledKeysToCallback;
    }
    /*
    @Override
	public int getPreferredWidth() {
		return textRenderer.computeTextWidth();
	}
    
    @Override
    public int getPreferredHeight() {
    	return editor.getPreferredHeight();
    }*/

    protected void doCallback(int key) {
        if(callbacks != null) {
            for(Callback cb : callbacks) {
                cb.callback(key);
            }
        }
    }

    public int getColumns() {
        return columns;
    }

    /**
     * This is used to determine the desired width of the EditField based on
     * it's font and the character 'X'
     * 
     * @param columns number of characters
     * @throws IllegalArgumentException if columns < 0
     */
    public void setColumns(int columns) {
        if(columns < 0) {
            throw new IllegalArgumentException("columns");
        }
        this.columns = columns;
    }

    /**
     * Set a new text for this EditField.
     * If the new text is longer then {@link #getMaxTextLength()} then it is truncated.
     * The selection is cleared.
     * The cursor is positioned at the end of the new text.
     * If a model is set, then the model is also updated.
     *
     * @param text the new text
     * @throws NullPointerException if text is null
     */
    public void setText(String text) {
        text = TextUtil.limitStringLength(text, maxTextLength);
        editBuffer.replace(0, editBuffer.length(), text);
        textRenderer.setCharSequence(text);
    }

    public String getText() {
        return editBuffer.toString();
    }

    public int getTextLength() {
        return editBuffer.length();
    }

    public int getMaxTextLength() {
        return maxTextLength;
    }

    public void setMaxTextLength(int maxTextLength) {
        this.maxTextLength = maxTextLength;
    }

    @Override
    protected void applyTheme(ThemeInfo themeInfo) {
        super.applyTheme(themeInfo);
        applyThemeEditField(themeInfo);
        setLineCount(linecount, scrollY);
    }

    protected void applyThemeEditField(ThemeInfo themeInfo) {
        selectionImage = themeInfo.getImage("selection");
        columns = themeInfo.getParameter("columns", 5);
    }

    @Override
    protected void layout() {
        layoutChildFullInnerArea(textRenderer);
        checkTextWidth();
        layoutInfoWindows();
    }

    @Override
    protected void positionChanged() {
        layoutInfoWindows();
    }

    private void layoutInfoWindows() {
        if(errorInfoWindow != null) {
            layoutErrorInfoWindow();
        }
    }

    private int computeInnerWidth() {
        int columns = Integer.toString(linecount).length() + 1;
        if (this.columns > columns) columns = this.columns;
        Font font = getFont();
        if(font != null) {
           return font.computeTextWidth("X")*columns;
        }
        return 0;
    }

    private int computeInnerHeight() {
        int lineHeight = getLineHeight();
        return lineHeight * numberOfLines;
    }

    @Override
    public int getMinWidth() {
        int minWidth = super.getMinWidth();
        minWidth = Math.max(minWidth, computeInnerWidth() + getBorderHorizontal());
        return minWidth;
    }

    @Override
    public int getMinHeight() {
        int minHeight = super.getMinHeight();
        minHeight = Math.max(minHeight, computeInnerHeight() + getBorderVertical());
        return minHeight;
    }

    @Override
    public int getPreferredInnerWidth() {
        return computeInnerWidth();
    }

    @Override
    public int getPreferredInnerHeight() {
        return computeInnerHeight();
    }

    public void setErrorMessage(Object errorMsg) {
        getAnimationState().setAnimationState(STATE_ERROR, errorMsg != null);
        if(this.errorMsg != errorMsg) {
            this.errorMsg = errorMsg;
            updateTooltip();
        }
        if(errorMsg != null) {
            if(hasKeyboardFocus()) {
                openErrorInfoWindow();
            }
        } else if(errorInfoWindow != null) {
            errorInfoWindow.closeInfo();
        }
    }

    @Override
    public Object getTooltipContent() {
        if(errorMsg != null) {
            return errorMsg;
        }
        Object tooltip = super.getTooltipContent();
        if(tooltip == null && textLongerThenWidget && !hasKeyboardFocus()) {
            tooltip = "Double click to set a breakpoint";//getText();
        }
        return tooltip;
    }

    @Override
    public boolean handleEvent(Event evt) {

        if(evt.isMouseEvent()) {
            boolean hover = (evt.getType() != Event.Type.MOUSE_EXITED) && isMouseInside(evt);
            getAnimationState().setAnimationState(STATE_HOVER, hover);
        }

        if(evt.isMouseDragEvent()) {
            if(evt.getType() == Event.Type.MOUSE_DRAGGED &&
                    (evt.getModifiers() & Event.MODIFIER_LBUTTON) != 0) {
            }
            return true;
        }

        if(super.handleEvent(evt)) {
            return true;
        }
        
        switch (evt.getType()) {

        case MOUSE_BTNUP:
            break;

        case MOUSE_BTNDOWN:
            if(evt.getMouseButton() == Event.MOUSE_LBUTTON && isMouseInside(evt)) {
                return true;
            }
            break;

        case MOUSE_CLICKED:
            if(evt.getMouseClickCount() == 2) {
            	int lineNumber = getLineNumberFromMouse(evt.getMouseX(), evt.getMouseY()) - 1;
            	setLocalBreakPoint(lineNumber);
                return true;
            }
            if(evt.getMouseClickCount() == 3) {
                return true;
            }
            break;

        case MOUSE_WHEEL:
            return false;
        }

        return evt.isMouseEvent();
    }

    private void checkTextWidth() {
        textLongerThenWidget = textRenderer.getPreferredWidth() > textRenderer.getWidth();
    }

    protected boolean hasFocusOrPopup() {
        return hasKeyboardFocus() || hasOpenPopups();
    }
    
    protected Font getFont() {
        return textRenderer.getFont();
    }

    protected int getLineHeight() {
        Font font = getFont();
        if(font != null) {
            return font.getLineHeight();
        }
        return 0;
    }

    protected int computeLineNumber(int cursorPos) {
        final EditFieldModel eb = this.editBuffer;
        int lineNr = 0;
        for(int i=0 ; i<cursorPos ; i++) {
            if(eb.charAt(i) == '\n') {
                lineNr++;
            }
        }
        return lineNr;
    }

    protected int computeLineStart(int cursorPos) {
        final EditFieldModel eb = this.editBuffer;
        while(cursorPos > 0 && eb.charAt(cursorPos-1) != '\n') {
            cursorPos--;
        }
        return cursorPos;
    }

    protected int computeLineEnd(int cursorPos) {
        final EditFieldModel eb = this.editBuffer;
        int endIndex = eb.length();
        while(cursorPos < endIndex && eb.charAt(cursorPos) != '\n') {
            cursorPos++;
        }
        return cursorPos;
    }

    protected int computeRelativeCursorPositionX(int cursorPos) {
        int lineStart = 0;
        lineStart = computeLineStart(cursorPos);
        return textRenderer.computeRelativeCursorPositionX(lineStart, cursorPos);
    }

    protected int computeRelativeCursorPositionY(int cursorPos) {
            return getLineHeight() * computeLineNumber(cursorPos);
    }

    protected int getCursorPosFromMouse(int x, int y) {
        Font font = getFont();
        if(font != null) {
            int lineStart = 0;
            int lineEnd = editBuffer.length();
                y -= textRenderer.computeTextY();
                int lineHeight = font.getLineHeight();
                int endIndex = lineEnd;
                for(;;) {
                    lineEnd = computeLineEnd(lineStart);

                    if(lineStart >= endIndex || y < lineHeight) {
                        break;
                    }

                    lineStart = Math.min(lineEnd + 1, endIndex);
                    y -= lineHeight;
                }
            return computeCursorPosFromX(x, lineStart, lineEnd);
        } else {
            return 0;
        }
    }
    
    protected int getLineNumberFromMouse(int x, int y) {
    	Font font = getFont();
    	if (font != null) {
    		y -= textRenderer.computeTextY();
            int lineHeight = font.getLineHeight();
            int lineNumber = y / lineHeight + 1;
            if (linecount >= lineNumber) return lineNumber;
    	}
    	return 0;
    }

    protected int computeCursorPosFromX(int x, int lineStart) {
        return computeCursorPosFromX(x, lineStart, computeLineEnd(lineStart));
    }
    
    protected int computeCursorPosFromX(int x, int lineStart, int lineEnd) {
        Font font = getFont();
        if(font != null) {
            return lineStart + font.computeVisibleGlpyhs(editBuffer, lineStart, lineEnd, x + font.getSpaceWidth() / 2);
        }
        return lineStart;
    }
    
    @Override
    protected void paintOverlay(GUI gui) {
        super.paintOverlay(gui);
    }

    private void openErrorInfoWindow() {
            if(errorInfoWindow == null) {
                errorInfoLabel = new Label();
                errorInfoLabel.setClip(true);
                errorInfoWindow = new InfoWindow(this);
                errorInfoWindow.setTheme("editfield-errorinfowindow");
                errorInfoWindow.add(errorInfoLabel);
            }
            errorInfoLabel.setText(errorMsg.toString());
            errorInfoWindow.openInfo();
            layoutErrorInfoWindow();
    }

    private void layoutErrorInfoWindow() {
        errorInfoWindow.setSize(getWidth(), errorInfoWindow.getPreferredHeight());
        errorInfoWindow.setPosition(getX(), getBottom());
    }

    @Override
    protected void keyboardFocusGained() {
        if(errorMsg != null) {
            openErrorInfoWindow();
        }
    }

    @Override
    protected void keyboardFocusLost() {
        super.keyboardFocusLost();
        if(errorInfoWindow != null) {
            errorInfoWindow.closeInfo();
        }
    }

    protected class TextRenderer extends TextWidget {
        private int lastTextX;

		protected TextRenderer(AnimationState animState) {
            super(animState);
        }
        
        protected void setCharSequence(CharSequence text) {
        	super.setCharSequence(text);
        }
        
        protected int computeTextY() {
        	return scrollY;
        }

        @Override
        protected void paintWidget(GUI gui) {
        	final EditFieldModel eb = editBuffer;
        	lastTextX = computeTextX();
            int lineStart = 0;
            int endIndex = eb.length();
            int yoff = computeTextY();
            int lineHeight = getLineHeight();
            int linen = 0;
            boolean selected = false;
            while(lineStart < endIndex) {
                int lineEnd = computeLineEnd(lineStart);
                try {
                	if (breakpointLines[linen] == 2) selected = true;
                } catch (ArrayIndexOutOfBoundsException e) {
            	}
                paintWithSelection(lineStart, lineEnd, yoff, selected);
                linen++;
                yoff += lineHeight;
                lineStart = lineEnd + 1;
                selected = false;
            }
        }
        
        protected void paintWithSelection(int lineStart, int lineEnd, int yoff, boolean selected) {
            if(selectionImage != null && selected) {
                int xpos0 = lastTextX + computeRelativeCursorPositionX(lineStart, lineStart);
                int xpos1 = lastTextX + computeRelativeCursorPositionX(lineStart, lineEnd);
                selectionImage.draw(getAnimationState(), xpos0, yoff,
                        xpos1 - xpos0, getFont().getLineHeight());
                paintWithSelection(getAnimationState(), lineStart, lineEnd, lineStart, lineEnd, yoff);
            } else
            	paintWithSelection(getAnimationState(), 0, 0, lineStart, lineEnd, yoff);
        }

        @Override
        protected int computeTextX() {
            return getInnerX();// - lastScrollPos;
        }
        
        @Override
        public int computeTextWidth() {
        	if(getFont() != null) {
        		return getFont().computeMultiLineTextWidth(getText()) + 20;
        	}
        	return 0;
        }
    }

	private int linecount = 1;
	private int scrollY = 0;

	@Override
	public void setLineCount(int n, int scrollY) {
		this.scrollY = scrollY;
		if (n != linecount) {
			int width = Integer.toString(n).length();
			width = (width > columns) ? width : columns;
			String newText = "", tmp;
			for (int i = 0; i < n; i++) {
				tmp = Integer.toString(i+1);
				while (tmp.length() < width) tmp = " " + tmp;
				newText += tmp + "\n";
			}
			setText(newText); 
			linecount = n;
		}
		//updateScroll();
		invalidateLayoutLocally();
	}

	private int[] breakpointLines = new int[1];
	
	@Override
	public void setMeaningfullLines(int[] lines) {
		for (int i = 0; i < lines.length && i < breakpointLines.length; i++) {
			lines[i] = (lines[i] == 1 && breakpointLines[i] == 2) ? 2 : lines[i];
		}
		breakpointLines = lines;
	}
	
	public List<Integer> getBreakpoints() {
		List<Integer> result = new ArrayList<Integer>();
		for (int i = 0; i < breakpointLines.length; i++)
			if (breakpointLines[i] == 2) result.add(i+1);
		return result;
	}

	@Override
	public void setBreakPoint(int line) {
		setLocalBreakPoint(line);
	}
	
	protected void setLocalBreakPoint(int line) {
		if (line != -1) try {
    		if (breakpointLines[line] == 2) breakpointLines[line] = 1;
    		else if (breakpointLines[line] == 1) breakpointLines[line] = 2;
    	} catch (ArrayIndexOutOfBoundsException e) {
    	}
	}
}
