package org.ntnu.pythia.engine.components;

import java.util.concurrent.DelayQueue;
import java.util.concurrent.Delayed;
import java.util.concurrent.TimeUnit;

import org.ntnu.pythia.engine.App;
import org.ntnu.pythia.engine.Position;
import org.ntnu.pythia.game.PlayerObject;

import de.matthiasmann.twl.DesktopArea;
import de.matthiasmann.twl.Event;
import de.matthiasmann.twl.GUI;
import de.matthiasmann.twl.Label;
import de.matthiasmann.twl.ResizableFrame.ResizableAxis;
import de.matthiasmann.twl.Widget;

public class RootPane extends Widget {
	public final DesktopArea desk;
	public final HelpPanel helpPanel;
	public final StatusLine statusLine;
	public final ConsoleDialog consoleDialog;
	public final CodeEditorDialog codeEditorDialog;
	public final HealthBar healthBar;
	private InputEventListener eventListener;
	
	public void hideAll() {
		helpPanel.setVisible(false);
		statusLine.setVisible(false);
		consoleDialog.setVisible(false);
		codeEditorDialog.setVisible(false);
		healthBar.setVisible(false);
	}
	
	public void showHUD() {
		helpPanel.setVisible(true);
		statusLine.setVisible(true);
		healthBar.setVisible(true);
		updateHelpPanel();
	}
	
	private void updateHelpPanel() {
		int result = 0;
		if (consoleDialog.isVisible())
			result |= HelpPanel.WIDGET_CONSOLE;
		if (codeEditorDialog.isVisible())
			result |= HelpPanel.WIDGET_EDITOR;
		helpPanel.updateHelpPanel(result);
	}
	
	protected class DialogLabel extends Label implements Delayed {
		private final long delay;
		private final long startTime;
		private final Position camera;
		private final Position object;
		private final int dispY;
		public DialogLabel(String text, long delay, Position camera, Position object, int dispY) {
			super(text);
			this.delay = delay;
			this.camera = camera;
			this.object = object;
			this.dispY = dispY;
			startTime = System.currentTimeMillis();
		}
		@Override
		public void paint(GUI gui) {
			if (gui != null && camera != null && object != null) {
				int x = gui.getWidth() / 2 - camera.getX() + object.getX();
				int y = gui.getHeight() / 2 - camera.getY() + object.getY() - dispY - getFont().getLineHeight()/2;
				setPosition(x, y);
			}
			super.paint(gui);
		}
		private long remaining() {
			long remaining = delay - (System.currentTimeMillis() - startTime);
			return (remaining > 0) ? remaining : 0; 
		}
		@Override
		public int compareTo(Delayed o) {
			if (o == null) return 0;
			return (int)(remaining() - o.getDelay(TimeUnit.MILLISECONDS));
		}
		@Override
		public long getDelay(TimeUnit unit) {
			return unit.convert(remaining(), TimeUnit.MILLISECONDS);
		}
	}
	
	private final DelayQueue<DialogLabel> dialogs = new DelayQueue<DialogLabel>();

    public RootPane() {
        setTheme("");
        setFocusKeyEnabled(false);
        
        helpPanel = new HelpPanel();
        consoleDialog = new ConsoleDialog();
        codeEditorDialog = new CodeEditorDialog();
        statusLine = new StatusLine();
        healthBar = new HealthBar();
        healthBar.setTheme("/progressbar");
        
        desk = new DesktopArea();
        desk.setTheme("");

        add(desk);
        
        desk.add(consoleDialog);
        consoleDialog.adjustSize();
        consoleDialog.setVisible(false);
        consoleDialog.setResizableAxis(ResizableAxis.NONE);
        desk.add(codeEditorDialog);
        codeEditorDialog.adjustSize();
        codeEditorDialog.setVisible(false);
        codeEditorDialog.setResizableAxis(ResizableAxis.NONE);
        desk.add(statusLine);
        statusLine.adjustSize();
        statusLine.setVisible(true);
        statusLine.setResizableAxis(ResizableAxis.NONE);
        desk.add(helpPanel);
        helpPanel.adjustSize();
        helpPanel.setVisible(true);
        helpPanel.setResizableAxis(ResizableAxis.NONE);
        healthBar.setSize(200, 20);
        desk.add(healthBar);
        healthBar.setPosition(60, 60);
        healthBar.setVisible(false);
    }
    
    public void setPlayer(PlayerObject player) {
    	healthBar.setGameObject(player);
    }
    
    public void setEventListener(InputEventListener eventListener) {
		this.eventListener = eventListener;
	}
    
    @Override
    protected void keyboardFocusGained() {
    	if (consoleDialog.isVisible()) consoleDialog.requestKeyboardFocus();
    	else if (codeEditorDialog.isVisible()) codeEditorDialog.requestKeyboardFocus();
    }

	@Override
    protected void layout() {
        desk.setSize(getParent().getWidth(), getParent().getHeight());
    }

    @Override
    protected void afterAddToGUI(GUI gui) {
        super.afterAddToGUI(gui);
        validateLayout();
    }
    
    public void showConsole(String text) {
    	if (codeEditorDialog.isVisible()) codeEditorDialog.setVisible(false);
    	else consoleDialog.setVisible(true);
    	consoleDialog.output("\n" + text);
    	updateHelpPanel();
    }
    
    public void showConsole() {
    	if (codeEditorDialog.isVisible()) codeEditorDialog.setVisible(false);
    	else consoleDialog.setVisible(true);
    	consoleDialog.showLogo();
    	updateHelpPanel();
    }
    
    public void showEditor() {
    	if (consoleDialog.isVisible()) consoleDialog.setVisible(false);
    	if (codeEditorDialog.isVisible()) codeEditorDialog.setVisible(false);
    	else codeEditorDialog.setVisible(true);
    	updateHelpPanel();
    }
    
    public boolean gameHasFocus() {
    	return !(codeEditorDialog.isVisible() || consoleDialog.isVisible() || this.hasOpenPopups());
    }
    
    @Override
    protected boolean handleEvent(Event evt) {
        if (evt.getType() == Event.Type.KEY_PRESSED && evt.getKeyCode() == Event.KEY_F1) {
        	if (!consoleDialog.isVisible()) {if(App.globals.hasLaptop()) showConsole("");}
        	else consoleDialog.setVisible(false);
        	updateHelpPanel();
        	return true;
        } else if (evt.getType() == Event.Type.KEY_PRESSED && evt.getKeyCode() == Event.KEY_F2) {
        	if(App.globals.hasEditor()) showEditor();
        	return true;
        }
        
        if (!super.handleEvent(evt) && eventListener != null) return eventListener.handleEvent(evt);
    	return true;
    }
    
    @Override
    public void paint(GUI gui) {
    	DialogLabel lbl = dialogs.poll(); 
    	if (lbl != null) removeChild(lbl);
    	super.paint(gui);
    }
    
    public void showDialog(String text, long ms, Position camera, Position object, int dispY) {
    	DialogLabel label = new DialogLabel(text, ms, camera, object, dispY);
    	label.setTheme("/dialoglabel");
    	add(label);
    	dialogs.add(label);
    }
}
