package org.ntnu.pythia.engine.lwjgl;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.color.ColorSpace;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.ComponentColorModel;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferByte;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;
import java.io.BufferedInputStream;
import java.io.File;
import java.io.IOException;
import java.net.URL;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.IntBuffer;
import java.util.HashMap;
import java.util.Hashtable;

import javax.imageio.ImageIO;

import org.lwjgl.opengl.GL11;
import org.ntnu.pythia.engine.App;

public class TextureLoader {
	private HashMap<String, Texture> table = new HashMap<String, Texture>();
	private ColorModel glAlphaColorModel;
	private ColorModel glColorModel;
	
	public TextureLoader() {
		glAlphaColorModel = new ComponentColorModel(ColorSpace.getInstance(ColorSpace.CS_sRGB),
				new int[] {8, 8, 8, 8},
				true,
				false,
				ComponentColorModel.TRANSLUCENT,
				DataBuffer.TYPE_BYTE);
		glColorModel = new ComponentColorModel(ColorSpace.getInstance(ColorSpace.CS_sRGB),
				new int[] {8, 8, 8, 0},
				false,
				false,
				ComponentColorModel.OPAQUE,
				DataBuffer.TYPE_BYTE);
	}
	
	private int createTextureID() {
		IntBuffer tmp = createIntBuffer(1);
		GL11.glGenTextures(tmp);
		return tmp.get(0);
	}
	
	public Texture getTexture(String resourceName) throws IOException {
		Texture tex = (Texture) table.get(resourceName);
		
		if (tex != null) return tex;
		
		tex = getTexture(resourceName,
				GL11.GL_TEXTURE_2D,
				GL11.GL_RGBA,
				GL11.GL_LINEAR,
				GL11.GL_LINEAR);
		
		table.put(resourceName, tex);
		
		return tex;
	}
	
	private Texture getTexture(String resourceName, int target, int dstPixelFormat, int minFilter, int magFilter) throws IOException {
		int srcPixelFormat = 0;
		int textureID = createTextureID();
		Texture texture = new Texture(target, textureID);
		texture.bind();
		BufferedImage bufferedImage = loadImage(resourceName);
		if(bufferedImage == null)
			throw new RuntimeException("Missing image at location: '" +resourceName+ "'.");
		texture.setImageHeight(bufferedImage.getHeight());
		texture.setImageWidth(bufferedImage.getWidth());
		
		srcPixelFormat = (bufferedImage.getColorModel().hasAlpha()) ? GL11.GL_RGBA : GL11.GL_RGB;
		
		ByteBuffer textureBuffer = convertImageData(bufferedImage, texture);
		
		if (target == GL11.GL_TEXTURE_2D) {
			GL11.glTexParameteri(target, GL11.GL_TEXTURE_MIN_FILTER, minFilter);
			GL11.glTexParameteri(target, GL11.GL_TEXTURE_MAG_FILTER, magFilter);
		}
		
		GL11.glTexImage2D(target, 0, dstPixelFormat, get2Fold(bufferedImage.getWidth()), get2Fold(bufferedImage.getHeight()),
				0, srcPixelFormat, GL11.GL_UNSIGNED_BYTE, textureBuffer);
		return texture;
	}
	
	private int get2Fold(int fold) {
		int ret = 2;
		while (ret < fold)
			ret *= 2;
		return ret;
	}
	
	private ByteBuffer convertImageData(BufferedImage bufferedImage, Texture texture) {
		ByteBuffer imageBuffer = null;
		WritableRaster raster;
		BufferedImage texImage;
		
		int texWidth = get2Fold(bufferedImage.getWidth());
		int texHeight = get2Fold(bufferedImage.getHeight());
		
		texture.setTextureHeight(texHeight);
		texture.setTextureWidth(texWidth);
		
		if (bufferedImage.getColorModel().hasAlpha()) {
			raster = Raster.createInterleavedRaster(DataBuffer.TYPE_BYTE, texWidth, texHeight, 4, null);
			texImage = new BufferedImage(glAlphaColorModel, raster, false, new Hashtable<String, Object>());
		} else {
			raster = Raster.createInterleavedRaster(DataBuffer.TYPE_BYTE, texWidth, texHeight, 3, null);
			texImage = new BufferedImage(glColorModel, raster, false, new Hashtable<String, Object>());
		}
		
		Graphics g = texImage.getGraphics();
		g.setColor(new Color(0f, 0f, 0f, 0f));
		g.fillRect(0, 0, texWidth, texHeight);
		g.drawImage(bufferedImage, 0, 0, null);
		
		byte[] data = ((DataBufferByte) texImage.getRaster().getDataBuffer() ).getData();
		
		imageBuffer = ByteBuffer.allocateDirect(data.length);
		imageBuffer.order(ByteOrder.nativeOrder());
		imageBuffer.put(data, 0, data.length);
		imageBuffer.flip();
		
		return imageBuffer;
	}
	
	private BufferedImage loadImage(String ref) throws IOException {
		BufferedImage bufferedImage;
		if(App.options.isExternal()){
			try{
				bufferedImage = ImageIO.read(new File(App.options.getExternal(), ref));
			}catch (IOException e) {
				URL url = TextureLoader.class.getClassLoader().getResource(ref);
				if (url == null) throw new IOException("Cannot find resource: " + ref);
				bufferedImage = ImageIO.read(new BufferedInputStream(getClass().getClassLoader().getResourceAsStream(ref)));
			}
		}else{
			URL url = TextureLoader.class.getClassLoader().getResource(ref);
			if (url == null) throw new IOException("Cannot find resource: " + ref);
			bufferedImage = ImageIO.read(new BufferedInputStream(getClass().getClassLoader().getResourceAsStream(ref)));
		}
		
		return bufferedImage;
	}
	
	protected IntBuffer createIntBuffer(int size) {
		ByteBuffer tmp = ByteBuffer.allocateDirect(4 * size);
		tmp.order(ByteOrder.nativeOrder());
		return tmp.asIntBuffer();
	}

}
