package org.ntnu.pythia.game;

import java.awt.geom.Rectangle2D;

import org.ntnu.pythia.game.behaviour.BehavedObject;

/**
 * A collision object for game objects.
 * Represents simple box
 * @author egr
 *
 */
public class CollisionObject {
	public static final int COLLISION_NONE = 0;
	public static final int COLLISION_NORTH = 1;
	public static final int COLLISION_SOUTH = -1;
	public static final int COLLISION_EAST = 2;
	public static final int COLLISION_WEST = -2;
	
	public boolean movable;
	public boolean backfires;
	
	public final int width;
	public final int height;
	
	public final GameObject parent;
	
	public final double posXdisp;
	public final double posYdisp;
	
	private final Rectangle2D.Double model;
	
	private boolean hollow = false;
	
	public CollisionObject(int width, int height, double posXdisp, double posYdisp, GameObject parent, boolean movable, boolean backfires, boolean hollow) {
		this.width = width;
		this.height = height;
		this.parent = parent;
		this.posXdisp = posXdisp;
		this.posYdisp = posYdisp;
		this.movable = movable;
		this.backfires = backfires;
		this.hollow = hollow;
		model = new Rectangle2D.Double(0, 0, width, height);
	}
	
	/**
	 * Collision object constructor
	 * @param width of the collision's model
	 * @param height of the collision's model
	 * @param parent is the game object the collision model is for
	 */
	public CollisionObject(int width, int height, double posXdisp, double posYdisp, GameObject parent, boolean movable, boolean backfires) {
		this.width = width;
		this.height = height;
		this.parent = parent;
		this.posXdisp = posXdisp;
		this.posYdisp = posYdisp;
		this.movable = movable;
		this.backfires = backfires;
		model = new Rectangle2D.Double(0, 0, width, height);
	}
	
	public CollisionObject(int width, int height, double posXdisp, double posYdisp, GameObject parent, boolean movable) {
		this(width, height, posXdisp, posYdisp, parent, movable, true);
	}
	
	private double getPosX() {
		return parent.posX + posXdisp;
	}
	
	private double getPosY() {
		return parent.posY + posYdisp;
	}
	
	public Rectangle2D.Double getModel() {
		model.x = getPosX() - width / 2;
		model.y = getPosY() - height / 2;
		return model;
	}
	
	/**
	 * Check if this object collides with another object
	 * @param o object to check
	 * @return collision direction
	 */
	Rectangle2D result = new Rectangle2D.Double();
	public void collide(CollisionObject o) {
		if (o == null) return;
		
		int direction;
		Rectangle2D model = getModel();
		Rectangle2D target = o.getModel();
		
		Rectangle2D.intersect(model, target, result);
		if (result.isEmpty()) return;
		if (result.getHeight() > result.getWidth()) {
			if (target.getX() > model.getX() ) direction = COLLISION_EAST;
			else direction = COLLISION_WEST;
		} else {
			if (target.getY() > model.getY()) direction = COLLISION_SOUTH;
			else direction = COLLISION_NORTH;
		}
		
		
		if (parent.isHarmful() && o.parent.isMortal()) {
			float fraction;
			if (hollow) {
				fraction = (float)(result.getWidth() * result.getHeight() / (target.getWidth() * target.getHeight()));
				if (fraction > 0.49) {
					o.parent.fall(parent);
					if (o.parent instanceof BehavedObject) ((BehavedObject)o.parent).react("fall");
				}
			} else {
				if (parent.getHarmfulAreal()) {
					fraction = (float)(result.getWidth() * result.getHeight() / (model.getWidth() * model.getHeight()));
					o.parent.hurt(parent.getDamagePerSecond() * fraction);
				} else {o.parent.hurt(parent.getDamagePerSecond());}
				if (o.parent instanceof BehavedObject) ((BehavedObject)o.parent).react("hurt");
			}
		}
		if (o.parent.isHarmful() && parent.isMortal()) {
			float fraction;
			if (o.isHollow()){
				fraction = (float)(result.getWidth() * result.getHeight() / (model.getWidth() * model.getHeight()));
				if (fraction > 0.4) {
					parent.fall(o.parent);
					if (parent instanceof BehavedObject) ((BehavedObject)parent).react("fall");
				}
			} else {
				if (o.parent.getHarmfulAreal()) {
					fraction = (float)(result.getWidth() * result.getHeight() / (target.getWidth() * target.getHeight()));
					parent.hurt(o.parent.getDamagePerSecond() * fraction);
				} else {parent.hurt(o.parent.getDamagePerSecond());}
				if (parent instanceof BehavedObject) ((BehavedObject)parent).react("hurt");
			}
		}
			
		if (o.backfires) if (movable) parent.collide(direction);
		if (backfires) if (o.movable) o.parent.collide(-direction);
		if (o.backfires) if (parent instanceof BehavedObject) fireBehavedEvent((BehavedObject)parent, direction);
		if (backfires) if (o.parent instanceof BehavedObject) fireBehavedEvent((BehavedObject)o.parent, -direction);
	}
	
	private void fireBehavedEvent(BehavedObject object, int direction) {
		switch (direction) {
		case COLLISION_EAST:
			object.react("collision-east");
			break;
		case COLLISION_WEST:
			object.react("collision-west");
			break;
		case COLLISION_NORTH:
			object.react("collision-north");
			break;
		case COLLISION_SOUTH:
			object.react("collision-south");
			break;
		default:
			break;
		}
	}
	
	public boolean isHollow() {
		return hollow;
	}
}
