package org.ntnu.pythia.game;

import java.awt.Color;
import java.awt.geom.Point2D;
import java.util.HashMap;
import java.util.Map;

import org.ntnu.pythia.engine.GameWorld;
import org.ntnu.pythia.engine.Position;
import org.ntnu.pythia.engine.ResourceFactory;

/**
 * Everything in Game is a GameObject! And I say everything! No question!
 * @author egr
 *
 */
public abstract class GameObject implements Position {
	/*** Common properties ***/
	/** Object's name */
	private final String name;
	protected int width;
	protected int height;
	/** The object's x position */
	protected double posX;
	/** The object's y position */
	protected double posY;
	/** Z-coordinate*/
	protected int z_index;
	/** The object's rotation */
	protected double rot = 0.0f;
	/** The pointer to current sprite */
	protected RenderObject currentSprite;
	/** The store of the last sprites to optimize sprite loading */
	protected final Map<String, RenderObject> localSpriteStore = new HashMap<String, RenderObject>();
	/** The only link to the outside world */
	protected GameWorld gameWorld;
	/** The current collision model of the object */
	private CollisionObject model;
	// FIXME: remove maybe?
	protected Color color;
	
	/*** Movable objects ***/
	/** The x velocity of the object in pixels per second */
	protected double vX;
	/** The y velocity of the object in pixels per second */
	protected double vY;
	/** The speed limit **/
	protected double vLimit = 100;
	/** The acceleration **/
	protected double a = 100;
	/** The object's x position */
	protected double lastPosX;
	/** The object's y position */
	protected double lastPosY;
	/** Movement lock */
	protected boolean moveRight = false;
	protected boolean moveLeft = false;
	protected boolean moveUp = false;
	protected boolean moveDown = false;
	
	/** The reference to the current sprite */
	protected String appearance;
	
	/*** Mortality ***/
	private boolean mortal = false;
	private float health = 100f;
	private float health_reg_rate = 0.5f;
	private boolean die_on_collision;
	private RenderObject death_animation = null;
	private float damage = 0;
	protected boolean falling = false;
	private float falling_progress = 0;
	/*** Harmful ***/
	private boolean harmful = false;
	private boolean harmful_areal = false;
	private float damage_per_second = 10f;
	
	public void setMortal(boolean mortal, float health, float health_reg_rate, boolean die_on_collision, String death_animation_ref) {
		this.mortal = mortal;
		this.health = health;
		this.health_reg_rate = health_reg_rate;
		this.die_on_collision = die_on_collision;
		if (death_animation_ref != null) this.death_animation = ResourceFactory.getRenderObject(death_animation_ref);
		if (death_animation instanceof AnimationObject) ((AnimationObject)death_animation).setPlayonce(true);
	}
	
	public void setHarmful(boolean harmful, float damage_per_second, boolean harmful_areal) {
		this.harmful = harmful;
		this.damage_per_second = damage_per_second;
		this.harmful_areal = harmful_areal;
	}
	
	public boolean getHarmfulAreal() {
		return harmful_areal;
	}
	
	public float getDamagePerSecond() {
		return damage_per_second;
	}
	
	public boolean isMortal() {
		if (falling && falling_progress < 3.0) return false;
		return mortal;
	}
	
	public boolean isDead() {
		return (health < 0f || (falling && falling_progress > 3.0));
	}
	
	public void hurt(float damage) {
		this.damage += damage;
	}
	
	public boolean isHarmful() {
		return harmful;
	}
	
	public float getHealth() {
		return health;
	}
	
	/**
	 * 
	 * @param name
	 * @param x
	 * @param y
	 * @param z_index
	 * @param width
	 * @param height
	 */
	public GameObject(String name, int x, int y, int z_index, int width, int height) {
		this.name = name;
		this.posX = x;
		this.posY = y;
		this.lastPosX = posX;
		this.lastPosY = posY;
		this.width = width;
		this.height = height;
		this.z_index = z_index;
	}
	
	/**
	 * GameObject constructor
	 * @param color Color
	 * @param x X coordinate in pixels
	 * @param y Y coordinate in pixels
	 * @param width Width in pixels
	 * @param height Height in meters!
	 */
	public GameObject(int x, int y, int width, int height, int z_index, Color color) {
		this(null, x, y, z_index, width, height);
		currentSprite = ResourceFactory.getRenderObject(color);
		appearance ="<fill color>";
	}
	
	/**
	 * GameObject constructor
	 * @param refs String array with file paths
	 * @param x X coordinate in pixels
	 * @param y Y coordinate in pixels
	 * @param width Width in pixels
	 * @param height Height in meters!
	 */
	public GameObject(int x, int y, int width, int height, int z_index, String ref) {
		this(null, x, y, z_index, width, height);
		currentSprite = ResourceFactory.getRenderObject(ref);
		appearance = ref;
	}
	
	/**
	 * GameObject constructor
	 * @param color Color
	 * @param x X coordinate in pixels
	 * @param y Y coordinate in pixels
	 * @param width Width in pixels
	 * @param height Height in meters!
	 */
	public GameObject(String name, int x, int y, int z_index, int width, int height, Color color) {
		this(name, x, y, z_index, width, height);
		currentSprite = ResourceFactory.getRenderObject(color);
		appearance ="<fill color>";
	}
	
	/**
	 * GameObject constructor
	 * @param refs String array with file paths
	 * @param x X coordinate in pixels
	 * @param y Y coordinate in pixels
	 * @param width Width in pixels
	 * @param height Height in meters!
	 */
	public GameObject(String name, int x, int y, int z_index, int width, int height, String ref) {
		this(name, x, y, z_index, width, height);
		currentSprite = ResourceFactory.getRenderObject(ref);
		appearance = ref;
	}
	
	/**
	 * Sets current sprite by the filename reference
	 * Note! The sprite must be loaded on construction
	 * @param source
	 */
	public void setSprite(String ref){
		if (localSpriteStore.containsKey(ref))
			currentSprite = localSpriteStore.get(ref);
		else {
			if (localSpriteStore.size() > 16) localSpriteStore.clear();
			currentSprite = ResourceFactory.getRenderObject(ref);
			localSpriteStore.put(ref, currentSprite);
		}
	}
	
	/**
	 * Executed on death
	 */
	protected void die() {
		model = null;
		currentSprite = death_animation;
	}
	
	/**
	 * Executed on damage taken
	 */
	protected void hurting(double damage) {
		health -= damage;
		this.damage = 0;
	}
	
	/**
	 * Calculate a new position for the object when delta milliseconds has elapsed
	 * This does not work for STATIC type
	 * @param delta The amount of time since last update
	 */
	public void move(double delta) {
		if (falling) {
			if (falling_progress > 3.0) return;
			falling_progress += delta;
			return;
		}
		if (mortal) {
			if (health > 0) {
				if (health <= 100) health += health_reg_rate * delta;
				if (damage < 0) {
					if (health <= 100)
						hurting(damage * delta);
				} else if (damage > 0) {
					hurting(damage * delta);
					if (health < 0) {
						die();
						return;
					} 
				}
			} else {
				((AnimationObject)currentSprite).animate(delta);				
				return;
			}
			damage = 0;
		}
		if (currentSprite instanceof AnimationObject) ((AnimationObject)currentSprite).animate(delta);
		if (model != null && !model.movable) return;
		lastPosX = posX; lastPosY = posY;
		posX += delta * vX;
		posY += delta * vY;
		if (moveUp) {if (vY > 0) vY = 0; if (vY > -vLimit) vY -= a*delta; else vY = -vLimit;}
		if (moveDown) {if (vY < 0) vY = 0; if (vY < vLimit) vY += a*delta; else vY = vLimit;}
		if (!moveUp && !moveDown) vY = 0;
		if (moveLeft) {if (vX > 0) vX = 0; if (vX > -vLimit) vX -= a*delta; else vX = -vLimit;}
		if (moveRight) {if (vX < 0) vX = 0; if (vX < vLimit) vX += a*delta; else vX = vLimit;}
		if (!moveLeft && !moveRight) vX = 0;
	}
	
	/**
	 * Set to true if we want to start moving up
	 * @param value
	 */
	public void setMoveUp(boolean value) {
		moveUp = value;
	}
	
	/**
	 * Set to true if we want to start moving down
	 * @param value
	 */
	public void setMoveDown(boolean value) {
		moveDown = value;
	}
	
	/**
	 * Set to true if we want to start moving left
	 * @param value
	 */
	public void setMoveLeft(boolean value) {
		moveLeft = value;
	}
	
	/**
	 * Set to true if we want to start moving right
	 * @param value
	 */
	public void setMoveRight(boolean value) {
		moveRight = value;
	}
	
	/**
	 * Get the current X position
	 * @return
	 */
	public int getX() {
		return (int) posX;
	}
	
	/**
	 * Get the current Y position
	 * @return
	 */
	public int getY() {
		return (int) posY;
	}
	
	/**
	 * Set the X position
	 * @param x
	 */
	public void setX(int x){
		posX = x;
	}
	
	/**
	 * Set the Y position
	 * @param y
	 */
	public void setY(int y){
		posY = y;
	}
	
	/**
	 * Draw the game object on the screen.
	 * This one is called by the renderer and no one else!
	 */
	public void draw() {
		if (currentSprite != null) {
			if (falling) {
				double size_fraction = (3.0 - falling_progress) / 3.0;
				currentSprite.draw(posX, posY, falling_progress * 360.0 * 3, (int)(width * size_fraction), (int)(height * size_fraction));
				return;
			}
			if (z_index > 0) currentSprite.drawShadow(posX, posY, rot, width, height, z_index);
			currentSprite.draw(posX, posY, rot, width, height);
			if (mortal && damage > 0) currentSprite.drawDamage(posX, posY, rot, width, height, damage);
		}
		if (ResourceFactory.RENDER_COLLISION_MODELS && model != null) {
			RenderObject tst = ResourceFactory.getRenderObject(Color.BLUE);
			tst.draw((int) model.posXdisp + posX, model.posYdisp + posY, 0, (int)model.width, (int)model.height);
		}
	}

	/**
	 * Calculates distance between itself and another GameObject
	 * @param other
	 * @return
	 */
	public double distanceTo(GameObject other) {
		Point2D p1 = new Point2D.Double(this.posX, this.posY);
		Point2D p2 = new Point2D.Double(other.posX, other.posY);
		return (float) p1.distance(p2);
	}
	
	/**
	 * Assigns object to the GameWorld
	 * @param gameWorld
	 */
	public void setGameWorld(GameWorld gameWorld) {
		this.gameWorld = gameWorld;
	}
	
	/**
	 * Returns current collision model
	 * @return
	 */
	public CollisionObject getCollisionModel() {
		return model;
	}

	/**
	 * Sets custom collision model
	 * Note! This method will not added to the GameObject's collision set
	 * @param o
	 */
	public void setCollisionObject(CollisionObject model) {
		this.model = model;
	}
	
	
	
	public boolean getMoveRight() {
		return moveRight;
	}

	public boolean getMoveLeft() {
		return moveLeft;
	}

	public boolean getMoveUp() {
		return moveUp;
	}

	public boolean getMoveDown() {
		return moveDown;
	}
	
	public void say(String text) {
		if (gameWorld != null) gameWorld.say(this, text);
	}

	/**
	 * Update, all dynamic objects can now collide
	 * @param event
	 */
	public void collide(int direction) {
		if (die_on_collision) {
			health = -1;
		}
		switch (direction) {
		case CollisionObject.COLLISION_EAST:
			if (Math.abs(posX - lastPosX) < 0.00001) posX -= 2; // JUMP BACK
			else posX = lastPosX;
			vX = 0;
			break;
		case CollisionObject.COLLISION_WEST:
			if (Math.abs(posX - lastPosX) < 0.00001) posX += 2; // JUMP BACK
			else posX = lastPosX;
			vX = 0;
			break;
		case CollisionObject.COLLISION_SOUTH:
			if (Math.abs(posY - lastPosY) < 0.00001) posY -= 2; // JUMP BACK
			else posY = lastPosY;
			vY = 0;
			break;
		case CollisionObject.COLLISION_NORTH:
			if (Math.abs(posY - lastPosY) < 0.00001) posY += 2; // JUMP BACK
			else posY = lastPosY;
			vY = 0;
			break;
		default:
			break;
		}
	}

	public int getHeight() {
		return height;
	}
	public int getWidth() {
		return width;
	}

	public double getVx() {
		return vX;
	}

	public void setVx(double vX) {
		this.vX = vX;
	}

	public double getVy() {
		return vY;
	}

	public void setVy(double vY) {
		this.vY = vY;
	}
	
	public void resumeAnimation() {
		if (currentSprite != null) currentSprite.resume();
	}
	
	public void pauseAnimation() {
		if (currentSprite != null) currentSprite.pause();
	}
	
	public void messageBox(String text) {
		gameWorld.messageBox(text);
	}
	
	public String getName() {
		return name;
	}
	
	public boolean isSolid(){
		return (model != null && !model.getModel().isEmpty());
	}
	
	public String getAppearance() {
		return appearance;
	}
	public void setAppearance(String appearance) {
		this.appearance = appearance;
		if (appearance == "<fill color>") return;
		this.setSprite(appearance);
	}
	
	public void setAppearance(Color color, int width, int height) {
		this.appearance ="<fill color>";
		this.width = width;
		this.height = height;
	}
	
	public Color getColor() {
		return color;
	}
	public void setColor(Color color) {
		if(color == null) return;
		this.color = color;
		currentSprite = ResourceFactory.getRenderObject(color);
	}

	public int getZindex() {
		return z_index;
	}
	
	public void setZindex(int z_index) {
		this.z_index = z_index;
	}
	
	public SoundObject playSound(String ref) {
		SoundObject sound = ResourceFactory.getSoundObject(ref);
		if (sound != null) sound.play();
		return sound;
	}

	public void fall(GameObject parent) {
		model = null;
		falling = true;
		setX(parent.getX());
		setY(parent.getY());
	}
}
