package org.ntnu.pythia.game.behaviour;

import java.awt.Color;
import java.util.Map;

import org.ntnu.pythia.game.CollisionObject;
import org.ntnu.pythia.game.CompositeCollisionObject;
import org.ntnu.pythia.game.level.AbstractObjectFactory;
import org.ntnu.pythia.game.level.SimpleObject;
import org.ntnu.pythia.game.level.xml.CollisionScheme;
import org.ntnu.pythia.game.level.xml.Bind.Binding;
import org.ntnu.pythia.python.ListListener;
import org.ntnu.pythia.python.VariableListener;
import org.python.core.PyObject;

/**
 * A controller for several smaller objects. 
 * Will behave as a Manipulateable Object as well as having inner objects
 * which it can manipulate with a python script.
 * @author andreas
 *
 */
public class CompositeObject extends BehavedObject {
	
	public CompositeObject(String name, int xpos, int ypos, int z_index,
			int height, int width) {
		super(name, new Color(255, 0, 255, 0), xpos, ypos, z_index, width, height);
		CompositeCollisionObject cco = new CompositeCollisionObject(0, 0, 0, 0, this, true);
		setCollisionObject(cco);
		currentSprite = null;
	}

	protected int cellsHigh, cellsWide;
	
	protected Map<String, AbstractObjectFactory> templates;
	
	protected SimpleObject[] cells = new SimpleObject[0];
	protected String[] cellRefs = new String[0];

	private int cellheight;
	private int cellwidth;
	
	public void setCell(int index, String value){
		setCellAt(index%cellsWide, index / cellsWide, value);
	}
	
	public SimpleObject getCell(int index){
		return cells[index];
	}
	public Behaviour getBehaviour(){
		return behaviour;
	}
	
	public void setCells(SimpleObject[] cells){
		this.cells = cells;
		setX(getX());
		setY(getY());
	}
	
	/**
	 * Set the cell at the internal coordinates of the composite object according to the factory defined in the XML.
	 * @param x
	 * @param y
	 * @param value the name of the factory to be used to create the internal object at this cell.
	 */
	public void setCellAt(int x, int y, String value){
		if (cells == null) cells = new SimpleObject[cellsWide * cellsHigh];
			int index = x + cellsWide * y;
		if(index > cells.length)
			return;
		cellRefs[index] = value;
		SimpleObject cell = make(templates.get(value), x, y);
		if(cell instanceof CompositeObject)
			((CompositeObject) cell).rebuild();
		cells[index] = cell;
		cell.setX(getX() - (this.width / 2) + x * cellwidth -1);
		cell.setY(getY() - (this.height / 2) + y * cellheight -1);
	}
	
	private SimpleObject make(AbstractObjectFactory factory, int x, int y) {
		factory.setX(getX() - (this.width / 2) + x * cellwidth -1);
		factory.setWidth(cellwidth);
		factory.setY(getY() - (this.height / 2) + y * cellheight -1);
		factory.setHeight(cellheight);
		return factory.getObject();
	}
	
	private void rebuild(){
		for (int i = 0; i < cells.length; i++) {
			setCell(i, cellRefs[i]);
		}
	}
	
	@Override
	public void setX(int x) {
		super.setX(x);
		if (cells != null)
			for(int i = 0; i < cells.length; i++) {
				if (cells[i]!=null) cells[i].setX(x - (width / 2) + (i%cellsWide) * cellwidth-1);
				
			}
	}
	
	@Override
	public void setY(int y) {
		super.setY(y);
		if (cells != null)
			for(int i = 0; i < cells.length; i++) {
				if (cells[i]!=null) cells[i].setY(y - (height / 2) + (i/cellsWide) * cellheight-1);
			}
	}
	
	@Override
	public void setCollisionObject(CollisionObject o) {
		if (o instanceof CompositeCollisionObject)
			super.setCollisionObject(o);
	}
	
	/**
	 * Removes an internal object.
	 * @param x
	 * @param y
	 */
	public void removeCellAt(int x, int y){
		int index = x + cellsWide * y;
		if(index > cells.length)
			return;
		cells[index] = null;
	}
	
	/**
	 * 
	 * @return all the internal objects as an array.
	 */
	final public SimpleObject[] getCells() {
		return cells;
	}
	
	@Override
	public CollisionScheme getCScheme() {
		return null;
	}
		
	@Override
	public void bindVariableToAttribute(String variable, Binding bind, String defaultVal){
		if(bind.equals(Binding.Custom)){
			parser.addListener(variable, new VariableListener(){
				@Override
				public void variableSet(String name, PyObject newValue) {
					CompositeObject.this.behaviour.react(newValue.toString(), CompositeObject.this);
					CompositeObject.this.behaviour.varSet(name, newValue);
					for(SimpleObject o : cells){
						if (o instanceof BehavedObject) {
							BehavedObject bo = (BehavedObject) o;
							bo.behaviour.react(newValue.toString(), bo);
						}
					}
				}
			});
			if(defaultVal != null)
				parser.exec(variable + " = " + defaultVal);
		}else if(bind.equals(Binding.CustomList)){
			parser.addListener(variable, new ListListener() {
				
				@Override
				public void variableSet(String name, PyObject index, PyObject newValue) {
					CompositeObject.this.behaviour.listSet(parser.eval(name), index.asIndex(), newValue);
					for(SimpleObject o : cells){
						if (o instanceof BehavedObject) {
							BehavedObject bo = (BehavedObject) o;
							bo.behaviour.listSet(parser.eval(name), index.asIndex(), newValue);
						}
					}
				}
			});
		} else
			super.bindVariableToAttribute(variable, bind, defaultVal);
	}

	public void setCellWidth(int cellwidth) {
		this.cellwidth = cellwidth;
	}

	public void setCellHeight(int cellheight) {
		this.cellheight = cellheight;
	}

	public void setTemplates(Map<String, AbstractObjectFactory> templatesByName) {
		this.templates = templatesByName;
	}

	public void setCellsWide(int cellsWide) {
		this.cellsWide = cellsWide;
		cells = new SimpleObject[cellsWide * cellsHigh];
		cellRefs = new String[cellsWide * cellsHigh];
	}

	public void setCellsHigh(int cellsHigh) {
		this.cellsHigh = cellsHigh;
		cells = new SimpleObject[cellsWide * cellsHigh];
		cellRefs = new String[cellsWide * cellsHigh];
	}
	
	public int getCellsWide() {
		return cellsWide;
	}

	public int getCellsHigh() {
		return cellsHigh;
	}

}

