package org.ntnu.pythia.game.level;

import java.awt.Color;
import java.util.List;

import javax.xml.bind.annotation.XmlAttribute;
import javax.xml.bind.annotation.XmlElementRef;
import javax.xml.bind.annotation.XmlEnum;
import javax.xml.bind.annotation.XmlEnumValue;
import javax.xml.bind.annotation.XmlRootElement;


/**
 * The Appearance of a manipulateable object. (What it looks like on screen) 
 * It consists of a set of textures and fill colours that define what it looks like in a given state.
 * @author andreas
 *
 */
@XmlRootElement(name = "appearance")
public class Appearance {
	
	@XmlAttribute(name = "initial")
	String appearance = "initial";
	
	public enum RenderMethod{
		/** Use texture to render */
		TEXTURE,
		/** Use fill colour to render */
		COLOR,
	}
	
	@XmlEnum
	public enum ColorChoice{
		@XmlEnumValue(value = "red")
		RED(Color.RED, "red"), 
		@XmlEnumValue(value = "blue")
		BLUE(Color.BLUE, "blue"),
		@XmlEnumValue(value = "green")
		GREEN(Color.GREEN, "green"), 
		@XmlEnumValue(value = "black")
		BLACK(Color.BLACK, "black"),
		@XmlEnumValue(value = "yellow")
		YELLOW(Color.YELLOW, "yellow"),
		@XmlEnumValue(value = "white")
		WHITE(Color.WHITE, "white"),
		@XmlEnumValue(value = "orange")
		ORANGE(Color.ORANGE, "orange"),
		@XmlEnumValue(value = "violet")
		VIOLET(new Color(0x50, 0, 0x7f), "violet"),
		@XmlEnumValue(value = "indigo")
		INDIGO(new Color(0, 0, 0xaf), "indigo"),
		@XmlEnumValue(value = "invisible")
		INVISIBLE(new Color(0, 0, 0, 0), "invisible");

				
		private Color c;
		private String name;
		private ColorChoice(Color c, String name){
			this.c = c;
			this.name = name;
		}
		public Color getColor(){
			return c;
		}
		public static ColorChoice byColor(Color color) {
			for(ColorChoice cc : ColorChoice.values())
				if( cc.c.equals(color))
					return cc;
			return BLACK;
		}
		public static ColorChoice byName(String color) {
			for(ColorChoice cc : ColorChoice.values())
				if( cc.name.equalsIgnoreCase(color))
					return cc;
			return BLACK;
		}
	}
	
	/**
	 * Returns all texture appearances
	 * @return
	 */
	public String getAppearance() {
		if (renderings.size() > 0)
			return renderings.get(0).getId();
		return null;
	}
	
	@XmlElementRef
	private List<Rendering> renderings;
	
	public static abstract class Rendering{
		RenderMethod method;
		
		public RenderMethod getMethod() {
			return method;
		}
		public abstract String getId();
	}
	
	@XmlRootElement(name = "color")
	public static class FillColor extends Rendering{
		@XmlAttribute(name = "color", required = true)
		private ColorChoice color;
		
		public FillColor(){
			method = RenderMethod.COLOR;
		}
		
		public Color getColor(){
			return color.getColor();
		}

		@Override
		public String getId() {
			return color.toString();
		}
	}

	@XmlRootElement(name = "texture")
	public static class Texture extends Rendering{
		@XmlAttribute(name = "id", required = true)
		protected String id;
		public Texture(){
			method = RenderMethod.TEXTURE;
		}
		public String getId(){
			return id;
		}
	}
	
	public void setState(String state){
		this.appearance = state;
	}
	
	public Rendering getRendering(){
		if(renderings.isEmpty())
			throw new RuntimeException("No valid renderings found");
		for(Rendering r : renderings)
			if(appearance.equals(r.getId()))
				return r;
		return renderings.get(0);
	}
}
