package org.ntnu.pythia.game.level;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.SchemaOutputResolver;
import javax.xml.bind.Unmarshaller;
import javax.xml.bind.annotation.XmlAttribute;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlElementRef;
import javax.xml.bind.annotation.XmlRootElement;
import javax.xml.transform.Result;
import javax.xml.transform.stream.StreamResult;

import org.ntnu.pythia.engine.App;
import org.ntnu.pythia.game.behaviour.BehavedObjectFactory;
import org.ntnu.pythia.game.behaviour.BehaviourProperty;
import org.ntnu.pythia.game.behaviour.CompositeObjectFactory;
import org.ntnu.pythia.game.behaviour.ConversationState;
import org.ntnu.pythia.game.behaviour.Interaction;
import org.ntnu.pythia.game.level.LevelGraphics.AnimationRef;
import org.ntnu.pythia.game.level.LevelGraphics.GraphicRef;
import org.ntnu.pythia.game.level.LevelGraphics.TextureRef;
import org.ntnu.pythia.game.level.xml.Bind;
import org.ntnu.pythia.game.level.xml.CollisionScheme;
import org.ntnu.pythia.game.level.xml.CollisionScheme.ModelType;
import org.ntnu.pythia.game.level.xml.LevelObjectWrapper;
import org.ntnu.pythia.game.level.xml.TileDefinition;
import org.ntnu.pythia.python.Module;
import org.ntnu.pythia.python.PythiaParser;
import org.ntnu.pythia.python.Script;
import org.ntnu.pythia.python.unit.ArrayTest;
import org.ntnu.pythia.python.unit.ImpliciteObjective;
import org.ntnu.pythia.python.unit.OutputTest;
import org.ntnu.pythia.python.unit.PyTestMaker;
import org.ntnu.pythia.python.unit.Test;
import org.ntnu.pythia.python.unit.TestProperty;
import org.ntnu.pythia.python.unit.VarTest;
import org.python.core.PyException;

/**
 * Used to load levels from file.
 * @author andreas
 *
 */
@XmlRootElement(name = "Level")
public class LevelFactory {
	private static final String rootFileName = "level.xml";

	@XmlAttribute(name = "number")
	private int levelNumber;

	@XmlAttribute(name = "name")
	private String levelName;

	@XmlElement(name="define")
	private String define;

	@XmlElement(name = "Layout", required = true)
	private LevelLayout layout;

	@XmlElement(name = "LevelGraphics", required = true)
	private LevelGraphics graphics;

	@XmlElement(name = "LevelSounds", required = true)
	private LevelSounds sounds;

	@XmlElement(name = "LevelObjects")
	private LevelObjectWrapper levelObjects; 

	@XmlElement(name = "WelcomeMessage")
	private Interaction interaction; 

	@XmlElement(name = "SpawnLocationX", required=true)
	private int spawnx;

	@XmlElement(name = "SpawnLocationY", required=true)
	private int spawny;

	@XmlElementRef
	private List<Module> levelModules;
	
	@XmlElementRef
	List<Test> tests;

	//end of XML-y stuff

	private PythiaParser masterParser;

	private Level level;

	private Map<String, AbstractObjectFactory> factoriesByRef = new HashMap<String, AbstractObjectFactory>();
	
	private Map<String, Module> modulesByName = new HashMap<String, Module>();

	protected void construct() {
		level = new Level();
	}

	@XmlElement(name="ExternalResource")
	private List<String> paths;

	protected void init(){
		if(paths != null)
			LoadableResourceLoader.load(paths);
		masterParser = App.globals.getParser();
		try{
			if(define!= null && !define.equals(""))
				masterParser.exec(define);
		}catch (PyException e) {
			throw new RuntimeException("Error in define: ", e);
		}
		if(levelModules!= null)
			for(Module m : levelModules)
				modulesByName.put(m.getName(), m);
		
		App.globals.setCurrentLevelModules(modulesByName);
		
		List<AbstractObjectFactory> factories = levelObjects.getList();

		List<AbstractObjectFactory> extending = new LinkedList<AbstractObjectFactory>();
		
		for(AbstractObjectFactory factory : factories){
			factory.setParser(masterParser);
			SimpleObject obj;
			if(factory.extend != null)
				extending.add(factory);
			else{
				try{
				obj = factory.getObject();
				level.addObject(obj.getName(), obj);
				}catch (Exception e) {
					throw new RuntimeException("Error while loading object '" + factory.getName() + "'.", e);
				}
			}
			if(factory.ref != null)
				factoriesByRef.put(factory.ref, factory);
		}
		
		for(AbstractObjectFactory factory : extending){
			SimpleObject obj = factory.getObject(factoriesByRef.get(factory.extend));
			level.addObject(obj.getName(), obj);
		}
	}

	protected void setValues(){
		level.setLevelNumber(levelNumber);
		level.setLevelNumber(levelName);
		level.setLevelGraphics(graphics);
		level.setLevelSounds(sounds);
		level.setLevelLayout(layout);
		level.setWelcomeMessage(interaction); 
		level.setSpawnx(spawnx);
		level.setSpawny(spawny);
		level.setTests(tests);
	}

	public Level getLevel() {
		construct();
		setValues();
		init();
		return level;
	}

	/**
	 * Load a level that is defined inside the jars /levels folder.
	 * @param levelName
	 * @return
	 * @throws JAXBException 
	 * @throws IOException 
	 */
	public static Level loadInternal(String levelName) throws JAXBException, IOException{

		InputStream levelStream = 
			LevelFactory.class.getResourceAsStream("/levels/" + levelName + "/" + rootFileName);

		if (levelStream == null)
			throw new RuntimeException("Error while opening internal file '" + levelName + "/" + rootFileName + "'.");

		JAXBContext context = JAXBContext.newInstance(LevelFactory.class, LevelLayout.class, LevelGraphics.class,
				Script.class, ManipulateableObjectFactory.class, Bind.class, LevelObjectWrapper.class, TileDefinition.class,
				TileDefinition.TileRef.class, GraphicRef.class, TextureRef.class, AnimationRef.class,
				BehavedObjectFactory.class, Appearance.class, Appearance.Rendering.class, Appearance.FillColor.class,
				Appearance.Texture.class, Interaction.class, ConversationState.class, Test.class, ArrayTest.class, 
				VarTest.class, OutputTest.class, ImpliciteObjective.class, PyTestMaker.class, CollisionScheme.class,
				ExitObjectFactory.class, BehaviourProperty.class, TestProperty.class, ModelType.class, CompositeObjectFactory.class,
				Module.class);
		Unmarshaller unmarshaller = context.createUnmarshaller();
		LevelFactory factory = (LevelFactory)unmarshaller.unmarshal(levelStream);

		if(factory.graphics == null)
			throw new RuntimeException("Error while opening level '" + levelName + "'. LevelGraphics not defined.");

		factory.graphics.loadResource();

		if (factory.sounds != null)
			factory.sounds.loadResource();

		if(factory.layout == null)
			throw new RuntimeException("Error while opening level '" + levelName + "'. Layout not found.");

		factory.layout.loadLayout();		

		levelStream.close();

		return factory.getLevel();
	}

	/**
	 * Load a level from anywhere (needs to match file-structure)
	 * @param levelName
	 * @return
	 * @throws Exception
	 */
	public static Level loadExternal(String path) throws JAXBException, IOException{
		InputStream levelStream = new FileInputStream(new File(path + "/" + rootFileName));

		if (levelStream == null)
			throw new RuntimeException("Error while opening internal file '" + path + "/" + rootFileName + "'.");

		JAXBContext context = JAXBContext.newInstance(LevelFactory.class, LevelLayout.class, LevelGraphics.class,
				Script.class, ManipulateableObjectFactory.class, Bind.class, LevelObjectWrapper.class, TileDefinition.class,
				TileDefinition.TileRef.class, GraphicRef.class, TextureRef.class, AnimationRef.class,
				BehavedObjectFactory.class, Appearance.class, Appearance.Rendering.class, Appearance.FillColor.class,
				Appearance.Texture.class, Interaction.class, ConversationState.class, Test.class, ArrayTest.class, 
				VarTest.class, OutputTest.class, ImpliciteObjective.class, PyTestMaker.class, CollisionScheme.class,
				ExitObjectFactory.class, BehaviourProperty.class, TestProperty.class, ModelType.class, CompositeObjectFactory.class,
				Module.class);
		Unmarshaller unmarshaller = context.createUnmarshaller();
		LevelFactory factory = (LevelFactory)unmarshaller.unmarshal(levelStream);

		if(factory.graphics == null)
			throw new RuntimeException("Error while opening level at '" + path + "'. LevelGraphics not defined.");

		factory.graphics.loadResource();

		if (factory.sounds != null)
			factory.sounds.loadResource();

		if(factory.layout == null)
			throw new RuntimeException("Error while opening level at '" + path + "'. Layout not found.");

		factory.layout.loadLayout();		

		levelStream.close();

		return factory.getLevel();
	}

	/**
	 * Creates a XSD for Level-XMLs
	 * @param args
	 * @throws JAXBException
	 * @throws IOException
	 */
	public static void main(String...args) throws JAXBException, IOException{
		JAXBContext context = JAXBContext.newInstance(LevelFactory.class, LevelLayout.class, LevelGraphics.class,
				Script.class, ManipulateableObjectFactory.class, Bind.class, LevelObjectWrapper.class, TileDefinition.class,
				TileDefinition.TileRef.class, GraphicRef.class, TextureRef.class, AnimationRef.class,
				BehavedObjectFactory.class, Appearance.class, Appearance.Rendering.class, Appearance.FillColor.class,
				Appearance.Texture.class, Interaction.class, ConversationState.class, Test.class, ArrayTest.class, 
				VarTest.class, OutputTest.class, ImpliciteObjective.class, PyTestMaker.class, CollisionScheme.class,
				ExitObjectFactory.class, BehaviourProperty.class, TestProperty.class, ModelType.class, CompositeObjectFactory.class, 
				Module.class);
		context.generateSchema(new SchemaOutputResolver() {
			@SuppressWarnings("deprecation")
			public Result createOutput(String namespaceUri, String suggestedFileName) throws IOException {
				File file;
				OutputStream out;
				file = new File("src/main/resources/schema.xsd");
				out = new FileOutputStream(file);

				System.out.println("writing to " + file.getAbsolutePath());

				StreamResult ss = new StreamResult(out);
				ss.setSystemId(file.toURL().toExternalForm());
				return ss;
			}

		});
	}
}
