package org.ntnu.pythia.python;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.bind.Unmarshaller;
import javax.xml.bind.annotation.XmlElement;
import javax.xml.bind.annotation.XmlRootElement;

import org.ntnu.pythia.engine.App;
import org.ntnu.pythia.engine.ResourceFactory;
import org.ntnu.pythia.game.behaviour.BehavedObject;
import org.ntnu.pythia.game.level.Level;
import org.ntnu.pythia.game.level.SimpleObject;
import org.ntnu.pythia.game.player.Player;

/**
 * Globally accessible game data and modules. Will be saved between games.
 * 
 * @author andreas
 *
 */
@XmlRootElement(name = "Global")
public class Globals {
	private final static String globalsFileName = "global.xml";

	@XmlElement(name = "Module")
	private List<Module> modules;

	private Map<String, Module> modulesByName;
	private Map<String, Module> currentLevelModules;

	//	/** Levels by name of levels */
	//	Map<String, Level> levels; //TODO: Possibly change to levelLoader objects (levelFactory)

	//XXX: save these?
	private Map<String, String> worldVariables = new HashMap<String, String>();

	@XmlElement
	private Player currentPlayer;

	private Level current;

	final PythiaParser masterParser = new PythiaParser();

	public PythiaParser getParser(){
		return masterParser;
	}

	public void init(){
		modulesByName = new HashMap<String, Module>();
		for(Module m : modules){
			modulesByName.put(m.name, m);
		}
		
		if(currentPlayer != null)
			currentPlayer.init();
		
		loadLibs();
	}
	
	public boolean isLoadable(){
		return currentPlayer != null;
	}

	public void loadLibs(){
		//Pythia standard library :)
		getModule("pylib").load(masterParser);
		if(App.options.isDebug())
			getModule("debug").load(masterParser);
	}
	
	/**
	 * Create a new player on level 1. This will also set the game to start on level 1.  
	 * @param name
	 */
	public void newPlayer(String name){
		currentPlayer = new Player();
		currentPlayer.setName(name);
		currentPlayer.setCurrentLevelName(App.options.getFirstLevel());
		save();
	}

	/** 
	 * Save to file.
	 */
	public void save(){
		currentPlayer.prepareToSave();
		File f = null;
		try{
			JAXBContext context = JAXBContext.newInstance(Module.class, Globals.class, Player.class);
			Marshaller marshaller = context.createMarshaller();
			//URL url = getClass().getResource("/" + globalsFileName);
			f = new File(ResourceFactory.CURRENT_FOLDER, globalsFileName);
			f.getParentFile().mkdirs();
			if(f.createNewFile())
				System.out.println("Created new file at " + f.getAbsolutePath());
							
			marshaller.marshal(this, f);
		}catch (IOException e) {
			throw new RuntimeException("Error while writing globals. path:" + f.getAbsolutePath(), e);
		} catch (JAXBException e) {
			throw new RuntimeException("Error while writing globals. path:" + f.getAbsolutePath(), e);
		}
	}

	public void addModule(Module m){
		modules.add(m);
		if(modulesByName!= null)
			modulesByName.put(m.name, m);
	}

	public Module getModule(String name){
		Module m = modulesByName.get(name);
		if(m == null){
			m = currentLevelModules.get(name);
			if(m == null)
				throw new RuntimeException("Unknown Module: '" + name + "'.");
		}
		return m;
	}

	public void setCurrentLevelModules(Map<String, Module> ms){
		currentLevelModules = ms;
	}
	
	/**
	 * Read global game data.
	 * @return
	 */
	public static Globals readGlobals() {
		Globals g = null;
		try{
			JAXBContext context = JAXBContext.newInstance(Module.class, Globals.class, Player.class);
			Unmarshaller unmarshaller = context.createUnmarshaller();
			File f = new File(ResourceFactory.CURRENT_FOLDER, globalsFileName);
			InputStream is;
			if (f.exists()) is  = new FileInputStream(f);
			else is = Globals.class.getResourceAsStream("/" + globalsFileName);
			if (is == null)
				throw new RuntimeException("Could not find globals at '" + globalsFileName + "'.");
			g = (Globals)unmarshaller.unmarshal(is);
			is.close();
			g.init();
		} catch (JAXBException e) {
			throw new RuntimeException("XML-Error while reading globals.", e);
		} catch (IOException e) {
			throw new RuntimeException("IO-Error while reading globals.", e);
		}
		return g;
	}

	public Map<String, Module> getModules(){
		return modulesByName;
	}

	public void setCurrentLevel(Level mLevel, String name) {
		current = mLevel;
		currentPlayer.setCurrentLevelName(name);
		ResourceFactory.getEditor().levelLoaded();
	}

	public void setVar(String path, String state){
		if(!path.contains("."))
			return;
		String first = path.substring(0, path.indexOf('.'));		
		String rest = path.substring(path.indexOf('.') + 1);		
		if (first.equals("states")){
			SimpleObject o = current.getObject(rest);
			if (o != null)
				o.setConvoState(state);
		}else if (first.equals("properties")){
			String part2 = rest.substring(0, rest.indexOf('.'));
			rest = rest.substring(rest.indexOf('.')+1);
			SimpleObject o = current.getObject(part2);
			if (o != null)
				if (o instanceof BehavedObject) {
					BehavedObject bo = (BehavedObject) o;
					bo.setProperty(rest, state);
				}
		}else if (first.equals("globals")){
			worldVariables.put(rest, state);
		}else if (first.equals("keys")){
			if("true".equalsIgnoreCase(state))
				currentPlayer.giveKey(rest);
		}else
			System.out.println("invalid var " + path);
	}

	public String getVar(String path){
		if(!path.contains("."))
			return "unknown";
		String first = path.substring(0, path.indexOf('.'));		
		String rest = path.substring(path.indexOf('.')+1);
		if (first.equals("states")){
			SimpleObject o = current.getObject(rest);
			if (o != null)
				return o.getConvoState();
		}
		if (first.equals("properties")){
			String part2 = rest.substring(0, rest.indexOf('.'));
			rest = rest.substring(rest.indexOf('.')+1);
			SimpleObject o = current.getObject(part2);
			if (o != null)
				if (o instanceof BehavedObject) {
					BehavedObject bo = (BehavedObject) o;
					return bo.getProperty(rest);
				}
		}
		if (first.equals("globals")){
			return worldVariables.get(rest);
		}
		if (first.equals("keys")){
			return currentPlayer.hasKey(rest)+ "";
		}
		return "unknown";
	}

	public Level getCurrentLevel(){
		return current;
	}

	/**
	 * Decides what the first level to be loaded is. Uses save data and command line options. Defaults to level1 if nothing else is present.
	 * @return
	 */
	public String getFirstLevelName() {
		if(App.options.getFirstLevel() != null && !App.options.getFirstLevel().equals(""))
			return App.options.getFirstLevel();
		if(currentPlayer.getCurrentLevelName() != null && ! currentPlayer.getCurrentLevelName().equals(""))
			return currentPlayer.getCurrentLevelName();
		return "level1";
	}

	public String getPlayerName() {
		return currentPlayer.getName();
	}

	public void savePlayerScript(Module script) {
		currentPlayer.saveScript(script);
		save();
	}
	
	public Module getPlayerScript(String name){
		return currentPlayer.getScript(name);
	}

	public boolean playerHasKey(String string) {
		return currentPlayer.hasKey(string);
	}
	public boolean hasLaptop(){
		if(App.options.isDebug())
			return true;
		return currentPlayer.hasLaptop();
	}
	
	public boolean hasEditor(){
		if(App.options.isDebug())
			return true;
		return currentPlayer.hasEditor();
	}

}